# -*- encoding:utf-8 -*-
"""
    全局环境配置模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os
import platform
import sys
import warnings
from enum import Enum
from os import path

import numpy as np
import pandas as pd

from ..CoreBu.ABuFixes import six

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""暂时支持windows和mac os，不是windows就是mac os（不使用Darwin做判断），linux下没有完整测试"""
g_is_mac_os = platform.system().lower().find("windows") < 0 and sys.platform != "win32"
"""python版本环境，是否python3"""
g_is_py3 = six.PY3
"""ipython，是否ipython运行环境"""
g_is_ipython = True
"""主进程pid，使用并行时由于ABuEnvProcess会拷贝主进程注册了的模块信息，所以可以用g_main_pid来判断是否在主进程"""
g_main_pid = os.getpid()

try:
    # noinspection PyUnresolvedReferences
    __IPYTHON__
except NameError:
    g_is_ipython = False

# noinspection PyBroadException
try:
    # noinspection PyUnresolvedReferences
    import psutil

    """有psutil，使用psutil.cpu_count计算cpu个数"""
    g_cpu_cnt = psutil.cpu_count(logical=True) * 1
except ImportError:
    if g_is_py3:
        # noinspection PyUnresolvedReferences
        g_cpu_cnt = os.cpu_count()
    else:
        import multiprocessing as mp

        g_cpu_cnt = mp.cpu_count()
except:
    # 获取cpu个数失败，默认4个
    g_cpu_cnt = 4

"""pandas忽略赋值警告"""
pd.options.mode.chained_assignment = None

"""numpy，pandas显示控制，默认开启"""
g_display_control = True
if g_display_control:
    # pandas DataFrame表格最大显示行数
    pd.options.display.max_rows = 20
    # pandas DataFrame表格最大显示列数
    pd.options.display.max_columns = 20
    # pandas精度浮点数显示4位
    pd.options.display.precision = 4
    # numpy精度浮点数显示4位，不使用科学计数法
    np.set_printoptions(precision=4, suppress=True)

"""忽略所有警告，默认关闭"""
g_ignore_all_warnings = False
"""忽略库警告，默认打开"""
g_ignore_lib_warnings = True
if g_ignore_lib_warnings:
    # noinspection PyBroadException
    try:
        import matplotlib

        matplotlib.warnings.filterwarnings('ignore')
        matplotlib.warnings.simplefilter('ignore')
        import sklearn

        sklearn.warnings.filterwarnings('ignore')
        sklearn.warnings.simplefilter('ignore')
    except:
        pass
if g_ignore_all_warnings:
    warnings.filterwarnings('ignore')
    warnings.simplefilter('ignore')

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 数据目录 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
"""
    abu 文件目录根目录
    windows应该使用磁盘空间比较充足的盘符，比如：d://, e:/, f:///

    eg:
    root_drive = 'd://'
    root_drive = 'e://'
    root_drive = 'f://'
"""
root_drive = path.expanduser('~')

"""abu数据缓存主目录文件夹"""
g_project_root = path.join(root_drive, 'abu')
"""abu数据文件夹 ~/abu/data"""
g_project_data_dir = path.join(g_project_root, 'data')
"""abu日志文件夹 ~/abu/log"""
g_project_log_dir = path.join(g_project_root, 'log')
"""abu数据库文件夹 ~/abu/db"""
g_project_db_dir = path.join(g_project_root, 'db')
"""abu缓存文件夹 ~/abu/cache"""
g_project_cache_dir = path.join(g_project_data_dir, 'cache')
"""abu项目数据主文件目录，即项目中的RomDataBu位置"""
g_project_rom_data_dir = path.join(path.dirname(path.abspath(path.realpath(__file__))), '../RomDataBu')

"""abu日志文件 ~/abu/log/info.log"""
g_project_log_info = path.join(g_project_log_dir, 'info.log')

"""hdf5做为金融时间序列存储的路径"""
g_project_kl_df_data = path.join(g_project_data_dir, 'df_kl.h5')

_p_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.path.pardir))
"""使用书中相同的沙盒数据环境，RomDataBu/df_kl.h5中内置的金融时间序列文件"""
g_project_kl_df_data_example = os.path.join(_p_dir, 'RomDataBu/df_kl.h5')

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 数据目录 end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊


#  ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 数据源 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ CrawlBu end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
"""
chrome 驱动
"""
g_crawl_chrome_driver = None


# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ CrawlBu start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊


# TODO 缩短 E_MARKET_SOURCE_bd－>BD
class EMarketSourceType(Enum):
    """
        数据源，当数据获取不可靠时，可尝试切换数据源，更可连接私有的数据源
    """
    """百度 a股，美股，港股"""
    E_MARKET_SOURCE_bd = 0
    """腾讯 a股，美股，港股"""
    E_MARKET_SOURCE_tx = 1
    """网易 a股，美股，港股"""
    E_MARKET_SOURCE_nt = 2
    """新浪 美股"""
    E_MARKET_SOURCE_sn_us = 3

    """新浪 国内期货"""
    E_MARKET_SOURCE_sn_futures = 100
    """新浪 国际期货"""
    E_MARKET_SOURCE_sn_futures_gb = 101

    """火币 比特币，莱特币"""
    E_MARKET_SOURCE_hb_tc = 200


"""默认设置数据源使用E_MARKET_SOURCE_bd"""
g_market_source = EMarketSourceType.E_MARKET_SOURCE_bd

"""自定义的私有数据源类，默认None"""
g_private_data_source = None


# TODO 缩短 E_MARKET_TARGET_US－>US
class EMarketTargetType(Enum):
    """
        交易品种类型，即市场类型，
        eg. 美股市场, A股市场, 港股市场, 国内期货市场,
            美股期权市场, TC币市场（比特币等
    """
    """美股市场"""
    E_MARKET_TARGET_US = 'us'
    """A股市场"""
    E_MARKET_TARGET_CN = 'hs'
    """港股市场"""
    E_MARKET_TARGET_HK = 'hk'

    """国内期货市场"""
    E_MARKET_TARGET_FUTURES_CN = 'futures_cn'
    """国际期货市场"""
    E_MARKET_TARGET_FUTURES_GLOBAL = 'futures_global'
    """美股期权市场"""
    E_MARKET_TARGET_OPTIONS_US = 'options_us'

    """TC币市场（比特币等）"""
    E_MARKET_TARGET_TC = 'tc'


class EMarketSubType(Enum):
    """
        子市场（交易所）类型定义
    """

    """美股纽交所NYSE"""
    US_N = 'NYSE'
    """美股纳斯达克NASDAQ"""
    US_OQ = 'NASDAQ'
    """美股粉单市场"""
    US_PINK = 'PINK'
    """美股OTCMKTS"""
    US_OTC = 'OTCMKTS'
    """美国证券交易所"""
    US_AMEX = 'AMEX'
    """未上市"""
    US_PREIPO = 'PREIPO'

    """港股hk"""
    HK = 'hk'

    """上证交易所sh"""
    SH = 'sh'
    """深圳交易所sz"""
    SZ = 'sz'

    """大连商品交易所DCE'"""
    DCE = 'DCE'
    """郑州商品交易所ZZCE'"""
    ZZCE = 'ZZCE'
    """上海期货交易所SHFE'"""
    SHFE = 'SHFE'

    """伦敦金属交易所"""
    LME = 'LME'
    """芝加哥商品交易所"""
    CBOT = 'CBOT'
    """纽约商品交易所"""
    NYMEX = 'NYMEX'

    """币类子市场COIN'"""
    COIN = 'COIN'


"""切换目标操作市场，美股，A股，港股，期货，比特币等，默认美股市场"""
g_market_target = EMarketTargetType.E_MARKET_TARGET_US

"""市场中1年交易日，默认250日"""
g_market_trade_year = 250
if g_market_target == EMarketTargetType.E_MARKET_TARGET_US:
    # 美股252天
    g_market_trade_year = 252
if g_market_target == EMarketTargetType.E_MARKET_TARGET_TC:
    # 默认设置币类每天都可以交易
    g_market_trade_year = 365


# TODO EMarketDataSplitMode移动到市场请求相关对应的模块中
class EMarketDataSplitMode(Enum):
    """
        ABuSymbolPd中请求参数，关于是否需要与基准数据对齐切割
    """
    """直接取出所有data，不切割，即外部需要切割"""
    E_DATA_SPLIT_UNDO = 0
    """内部根据start，end取切割data"""
    E_DATA_SPLIT_SE = 1


# TODO 缩短 E_DATA_FETCH_NORMAL－>NORMAL
class EMarketDataFetchMode(Enum):
    """
        金融时间数据获取模式
    """
    """普通模式，尽量从本地获取数据，本地数据不满足的情况下进行网络请求"""
    E_DATA_FETCH_NORMAL = 0
    """强制从本地获取数据，本地数据不满足的情况下，返回None"""
    E_DATA_FETCH_FORCE_LOCAL = 1
    """强制从网络获取数据，不管本地数据是否满足"""
    E_DATA_FETCH_FORCE_NET = 2


"""
    金融时间数据获取模式模块设置g_data_fetch_mode，默认为E_DATA_FETCH_NORMAL，实际上默认值建议
    为E_DATA_FETCH_FORCE_LOCAL，所有数据提前使用ABu.run_kl_update完成updtae，之后使用本地数据回测，
    原因：
    1. mac os 10.9 later 多进程 ＋ numpy有系统bug
    2. hdf5并行容易写坏文件
    3. 执行效率更高
    4. 分开数据获取与回测流程，更容易问题分析
"""
g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_NORMAL

"""是否开启ipython example 环境，默认关闭False"""
_g_enable_example_env_ipython = False


def enable_example_env_ipython(show_log=True, check_cn=True):
    """
    只为在ipython example 环境中运行与书中一样的数据，即读取RomDataBu/df_kl.h5下的数据

    初始内置在RomDataBu/df_kl.h5.zip下的数据只有zip压缩包，因为git上面的文件最好不要超过50m，
    内置测试数据，包括美股，a股，期货，比特币，港股数据初始化在df_kl_ext.h5.zip中，通过解压zip
    之后将测试数据为df_kl.h5
    show_log: 是否显示enable example env will only read RomDataBu/df_kl.h5
    check_cn: 是否检测运行环境有中文路径
    :return:
    """

    if not os.path.exists(g_project_kl_df_data_example):
        # 如果还没有进行解压，开始解压df_kl.h5.zip
        data_example_zip = os.path.join(_p_dir, 'RomDataBu/df_kl.h5.zip')
        try:
            from zipfile import ZipFile
            zip_h5 = ZipFile(data_example_zip, 'r')
            unzip_dir = os.path.join(_p_dir, 'RomDataBu/')
            for h5 in zip_h5.namelist():
                zip_h5.extract(h5, unzip_dir)
            zip_h5.close()
        except Exception as e:
            # 解压测试数据zip失败，就不开启测试数据模式了
            print('example env failed! e={}'.format(e))
            return

    global _g_enable_example_env_ipython, g_data_fetch_mode
    _g_enable_example_env_ipython = True
    g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL
    if check_cn:
        try:
            from ..UtilBu.ABuStrUtil import str_is_cn, to_unicode
            if str_is_cn(str(__file__)):
                # 检测到运行环境路径中含有中文，严重错误，将出错，使用中文警告
                msg = u'严重错误！当前运行环境下有中文路径，abu将无法正常运行！请不要使用中文路径名称, 当前环境为{}'.format(
                    to_unicode(str(__file__)))
                logging.info(msg)
                return
        except Exception as e:
            logging.exception(e)
    if show_log:
        logging.info('enable example env will only read RomDataBu/df_kl.h5')


def disable_example_env_ipython(show_log=True):
    """
    只为在ipython example 环境中运行与书中一样的数据。，即读取RomDataBu/df_kl.h5下的数据
    show_log: 是否显示disable example env
    """
    global _g_enable_example_env_ipython, g_data_fetch_mode
    _g_enable_example_env_ipython = False
    g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_NORMAL
    if show_log:
        logging.info('disable example env')


class EDataCacheType(Enum):
    """
        金融时间序列数据缓存类型
    """

    """读取及写入最快 但非固态硬盘写入慢，存贮空间需要大"""
    E_DATA_CACHE_HDF5 = 0
    """读取及写入最慢 但非固态硬盘写速度还可以，存贮空间需要小"""
    E_DATA_CACHE_CSV = 1
    """适合分布式扩展，存贮空间需要大"""
    E_DATA_CACHE_MONGODB = 2


# """默认金融时间序列数据缓存类型为HDF5，单机固态硬盘推荐HDF5，非固态硬盘使用CSV，否则量大后hdf5写入速度无法接受"""
# g_data_cache_type = EDataCacheType.E_DATA_CACHE_HDF5
"""对外版本由于用户电脑性能，存储空间且winodws用户，python2用户多，所以更改默认存储类型为csv"""
g_data_cache_type = EDataCacheType.E_DATA_CACHE_CSV

"""csv模式下的存储路径"""
g_project_kl_df_data_csv = path.join(g_project_data_dir, 'csv')

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 数据源 end   ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊特征快照切割 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
"""是否开启机器学习特征收集, 开启后速度会慢，默认关闭False"""
g_enable_ml_feature = False

"""是否开启买入订单前生成k线图快照，默认关闭False"""
g_enable_take_kl_snapshot = False

"""是否开启选股切割训练集股票数据与测试集股票数据，默认关闭False"""
g_enable_train_test_split = False

"""是否开启选股使用上一次切割完成的测试集股票数据，默认关闭False"""
g_enable_last_split_test = False

"""是否开启选股使用上一次切割完成的训练集股票数据，默认关闭False"""
g_enable_last_split_train = False

"""选股切割训练集股票数据与测试集股票数据切割参数n_folds，默认10"""
g_split_tt_n_folds = 10
# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊特征快照切割 end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊


# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊主裁 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
# TODO 内置ump的设置move到ABuUmpManager中

"""是否开启裁判拦截机制: 主裁deg，默认关闭False"""
g_enable_ump_main_deg_block = False
"""是否开启裁判拦截机制: 主裁jump，默认关闭False"""
g_enable_ump_main_jump_block = False
"""是否开启裁判拦截机制: 主裁price，默认关闭False"""
g_enable_ump_main_price_block = False
"""是否开启裁判拦截机制: 主裁wave，默认关闭False"""
g_enable_ump_main_wave_block = False
# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊主裁 end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊边裁 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊

"""是否开启裁判拦截机制: 边裁deg，默认关闭False"""
g_enable_ump_edge_deg_block = False

"""是否开启裁判拦截机制: 边裁price，默认关闭False"""
g_enable_ump_edge_price_block = False

"""是否开启裁判拦截机制: 边裁wave，默认关闭False"""
g_enable_ump_edge_wave_block = False
"""是否开启裁判拦截机制: 边裁full，默认关闭False"""
g_enable_ump_edge_full_block = False


# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊边裁 end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊


#  ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 日志 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
# TODO 将log抽出来从env中
def init_logging():
    """
    logging相关初始化工作，配置log级别，默认写入路径，输出格式
    """
    if g_is_ipython and not g_is_py3:
        """ipython在python2的一些版本需要reload logging模块，否则不显示log信息"""
        # noinspection PyUnresolvedReferences, PyCompatibility
        reload(logging)
        # pass

    if not os.path.exists(g_project_log_dir):
        # 创建log文件夹
        os.makedirs(g_project_log_dir)

    # 输出格式规范
    # file_handler = logging.FileHandler(g_project_log_info, 'a', 'utf-8')
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
                        datefmt='%a, %d %b %Y %H:%M:%S',
                        filename=g_project_log_info,
                        filemode='a'
                        # handlers=[file_handler]
                        )

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # 屏幕打印只显示message
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)


init_logging()
#  ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊ 日志 end ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊



# -*- encoding:utf-8 -*-
"""数据下载图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from ..WidgetBu.ABuWGBRunBase import WidgetTimeModeMixin

import logging
from collections import OrderedDict

import ipywidgets as widgets
from ..CoreBu.ABu import run_kl_update
from ..CoreBu import ABuEnv
from ..UtilBu import ABuProgress
from ..WidgetBu.ABuWGBase import WidgetBase, browser_down_csv_zip
from ..CoreBu.ABuEnv import EMarketTargetType, EMarketSourceType

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetUpdate(WidgetBase, WidgetTimeModeMixin):
    """数据下载图形可视化类"""

    def __init__(self):
        tm_box = self.init_time_mode_ui()
        # 修改为默认使用开始结束日期
        self.time_mode.value = 1
        # 修改开始结束日期时间字符串
        self.start.value = '2011-08-08'
        self.end.value = '2017-08-08'

        self.market = widgets.Dropdown(
            options=OrderedDict({u'美股': EMarketTargetType.E_MARKET_TARGET_US.value,
                                 u'A股': EMarketTargetType.E_MARKET_TARGET_CN.value,
                                 u'港股': EMarketTargetType.E_MARKET_TARGET_HK.value,
                                 u'国内期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_CN.value,
                                 u'国际期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL.value,
                                 u'数字货币': EMarketTargetType.E_MARKET_TARGET_TC.value}),
            value=ABuEnv.g_market_target.value,
            description=u'下载更新市场:',
        )
        self.market.observe(self.on_market_change, names='value')

        """数据源进行切换"""
        self.data_source_accordion = widgets.Accordion()
        date_source_dict_us = OrderedDict({
            u'腾讯数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_tx.value,
            u'百度数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_bd.value,
            u'新浪美股(美股)': EMarketSourceType.E_MARKET_SOURCE_sn_us.value,
            u'网易数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_nt.value,
        })

        date_source_dict_cn = OrderedDict({
            u'百度数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_bd.value,
            u'腾讯数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_tx.value,
            u'网易数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_nt.value
        })

        date_source_dict_hk = OrderedDict({
            u'网易数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_nt.value,
            u'腾讯数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_tx.value,
            u'百度数据源(美股，A股，港股)': EMarketSourceType.E_MARKET_SOURCE_bd.value
        })

        date_source_dict_futures_cn = {u'新浪国内期货(国内期货)': EMarketSourceType.E_MARKET_SOURCE_sn_futures.value}
        date_source_dict_futures_gb = {u'新浪国际期货(国际期货)': EMarketSourceType.E_MARKET_SOURCE_sn_futures_gb.value}
        date_source_dict_futures_tc = {u'火币网：比特币，莱特币': EMarketSourceType.E_MARKET_SOURCE_hb_tc.value}

        self.date_source_market_map = {
            EMarketTargetType.E_MARKET_TARGET_US.value: date_source_dict_us,
            EMarketTargetType.E_MARKET_TARGET_CN.value: date_source_dict_cn,
            EMarketTargetType.E_MARKET_TARGET_HK.value: date_source_dict_hk,
            EMarketTargetType.E_MARKET_TARGET_FUTURES_CN.value: date_source_dict_futures_cn,
            EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL.value: date_source_dict_futures_gb,
            EMarketTargetType.E_MARKET_TARGET_TC.value: date_source_dict_futures_tc}

        self.current_date_source_dict = self.date_source_market_map[self.market.value]
        self.date_source = widgets.RadioButtons(
            options=list(self.current_date_source_dict.keys()),
            value=list(self.current_date_source_dict.keys())[0],
            description=u'数据源:',
            disabled=False
        )
        self.date_source.observe(self.on_date_source_change, names='value')

        self.yun_down_bt = widgets.Button(description=u'美股|A股|港股|币类|期货6年日k', layout=widgets.Layout(width='50%'),
                                          button_style='info')
        self.yun_down_bt.on_click(self.run_yun_down)

        self.run_kl_update_bt = widgets.Button(description=u'从数据源下载更新', layout=widgets.Layout(width='50%'),
                                               button_style='danger')
        self.run_kl_update_bt.on_click(self.run_kl_update)
        description = widgets.Textarea(
            value=u'非沙盒数据，特别是回测交易多的情况下，比如全市场测试，回测前需要先将数据进行更新。\n'
                  u'建议直接从云盘下载入库完毕的数据库，不需要从各个数据源再一个一个的下载数据进行入库。\n'
                  u'abupy内置数据源都只是为用户学习使用，并不能保证数据一直通畅，而且如果用户很在乎数据质量，'
                  u'比如有些数据源会有前复权数据错误问题，有些数据源成交量不准确等问题，那么就需要接入用户自己的数据源。\n'
                  u'接入用户的数据源请阅读教程第19节中相关内容',
            disabled=False,
            layout=widgets.Layout(height='200px')
        )
        self.widget = widgets.VBox(
            [description, self.market, tm_box, self.date_source, self.yun_down_bt, self.run_kl_update_bt])

    # noinspection PyUnusedLocal
    def run_yun_down(self, bt):
        """打开浏览器csv zip地址准备开始下载"""
        # 清理之前的输出结果
        ABuProgress.clear_output()
        browser_down_csv_zip(open_browser=True)

    # noinspection PyUnusedLocal
    def run_kl_update(self, bt):
        """数据下载更新主接口"""
        n_folds = 2
        start = None
        end = None
        if not self.run_years.disabled:
            # 如果使用年回测模式
            n_folds = self.run_years.value
        if not self.start.disabled:
            # 使用开始回测日期
            start = self.start.value
        if not self.end.disabled:
            # 使用结束回测日期
            end = self.end.value
        market = ABuEnv.g_market_target

        logging.info(u'开始下载更新{}市场{}-{}:{}年, 数据源{}'.format(
            market.value, '' if start is None else start, '' if end is None else end, n_folds,
            ABuEnv.g_market_source.value))
        run_kl_update(start=start, end=end, n_folds=n_folds, market=market, n_jobs=10)
        logging.info(u'下载更新完成')

    def time_mode_str(self):
        """实现混入WidgetTimeModeMixin，声明时间模块代表下载更新"""
        return u'下载更新'

    def on_date_source_change(self, change):
        """数据源界面操作进行改变"""
        # 改变设置值
        ABuEnv.g_market_source = EMarketSourceType(self.current_date_source_dict[change['new']])

    def on_market_change(self, change):
        """切换大盘市场"""
        ABuEnv.g_market_target = EMarketTargetType(change['new'])

        self.current_date_source_dict = self.date_source_market_map[ABuEnv.g_market_target.value]

        options = list(self.current_date_source_dict.keys())
        self.date_source.options = options
        self.date_source.value = options[0]



# -*- encoding:utf-8 -*-
"""股票池选股ui界面"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import ipywidgets as widgets

from ..CoreBu.ABuEnv import EMarketTargetType
from ..WidgetBu.ABuWGBase import WidgetBase, WidgetSearchBox
from ..MarketBu import ABuMarket
from ..CoreBu import ABuEnv
from ..UtilBu.ABuStrUtil import to_unicode
from ..MarketBu.ABuSymbolStock import AbuSymbolCN, AbuSymbolUS, AbuSymbolHK
from ..MarketBu.ABuSymbolFutures import AbuFuturesCn, AbuFuturesGB

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetSymbolChoice(WidgetBase):
    """股票池选股ui界面"""

    # noinspection PyProtectedMember
    def __init__(self):
        """构建股票池选股ui界面"""
        label_layout = widgets.Layout(width='300px', align_items='stretch', justify_content='space-between')
        self.cs_tip = widgets.Label(value=u'如果股池为空，回测将使用大盘市场中所有股票', layout=label_layout)
        # 股票池多选框
        self.choice_symbols = widgets.SelectMultiple(
            description=u'股池:',
            disabled=False,
            layout=widgets.Layout(width='300px', align_items='stretch', justify_content='space-between')
        )
        self.choice_symbols.observe(self.choice_symbols_select, names='value')

        # 构建所有沙盒中的数据序列
        market_title = [u'美股', u'A股', u'港股', u'国内', u'国际', u'币类']
        us_seed_symbol = [to_unicode('{}:{}'.format(AbuSymbolUS()[symbol].co_name.values[0], symbol))
                          for symbol in ABuMarket.K_SAND_BOX_US]
        cn_seed_symbol = [to_unicode('{}:{}'.format(AbuSymbolCN()[symbol].co_name.values[0], symbol))
                          for symbol in ABuMarket.K_SAND_BOX_CN]
        hk_seed_symbol = [to_unicode('{}:{}'.format(AbuSymbolHK()[symbol].co_name.values[0], symbol))
                          for symbol in ABuMarket.K_SAND_BOX_HK]
        fcn_seed_symbol = [to_unicode('{}:{}'.format(AbuFuturesCn()[symbol]['product'].values[0], symbol))
                           for symbol in AbuFuturesCn().symbol]
        fgb_seed_symbol = [to_unicode('{}:{}'.format(AbuFuturesGB()[symbol]['product'].values[0], symbol))
                           for symbol in AbuFuturesGB().symbol]
        # 沙盒中的数据序列构建数据字典
        self.market_dict = {u'美股': us_seed_symbol,
                            u'A股': cn_seed_symbol,
                            u'港股': hk_seed_symbol,
                            u'国内': fcn_seed_symbol,
                            u'国际': fgb_seed_symbol,
                            u'币类': [u'比特币:btc', u'莱特币:ltc']}

        # 一个市场一个tab，tab中的symbol为沙盒中的symbol
        self.market_widget_tab = widgets.Tab()
        self.market_symbol_widget = []
        for title in market_title:
            market_symbol = widgets.SelectMultiple(
                options=self.market_dict[title],
                description=title,
                disabled=False
            )
            market_symbol.observe(self.on_already_select, names='value')
            self.market_symbol_widget.append(market_symbol)
        self.market_widget_tab.children = self.market_symbol_widget
        for ind, name in enumerate(market_title):
            self.market_widget_tab.set_title(ind, name)

        self.sc_box = WidgetSearchBox(self.on_already_select)()

        # 下拉选择标尺大盘
        self.market = widgets.Dropdown(
            options={u'美股': EMarketTargetType.E_MARKET_TARGET_US.value,
                     u'A股': EMarketTargetType.E_MARKET_TARGET_CN.value,
                     u'港股': EMarketTargetType.E_MARKET_TARGET_HK.value,
                     u'国内期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_CN.value,
                     u'国际期货': EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL.value,
                     u'数字货币': EMarketTargetType.E_MARKET_TARGET_TC.value},
            value=ABuEnv.g_market_target.value,
            description=u'大盘市场:',
        )
        self.market.observe(self.on_market_change, names='value')

        market_tip = widgets.Label(value=u'大盘市场设置只影响收益对比标尺', layout=label_layout)
        market_box = widgets.VBox([self.market, market_tip])

        self.widget = widgets.VBox([self.cs_tip, self.choice_symbols, self.market_widget_tab,
                                    self.sc_box, market_box])

    def on_market_change(self, change):
        """切换大盘市场"""
        ABuEnv.g_market_target = EMarketTargetType(change['new'])

    def on_already_select(self, select):
        """搜索框或者内置沙盒symbol中点击放入股票池"""
        st_symbol = [symbol.split(':')[1] if symbol.find(':') > 0
                     else symbol for symbol in list(select['new'])]
        # 更新股票池中原有的symbol序列
        self.choice_symbols.options = list(set(st_symbol + self.choice_symbols.options))

    def choice_symbols_select(self, select):
        """股票池中点击删除股票池中对应的symbol"""
        # print(select)
        # FIXME BUG 低版本ipywidgets下删除的不对
        self.choice_symbols.options = list(set(self.choice_symbols.options) - set(select['new']))



# -*- encoding:utf-8 -*-
"""封装常用的分析方式及流程模块"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import os

import matplotlib.cbook as cbook
import matplotlib.pyplot as plt
import numpy as np
from scipy import interp
from sklearn import metrics
from sklearn import tree
from sklearn.base import ClusterMixin, clone
from sklearn.metrics import roc_curve, auc

from ..CoreBu.ABuFixes import KFold, learning_curve
from ..UtilBu.ABuDTUtil import warnings_filter
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import range
from ..UtilBu.ABuFileUtil import file_exist

__author__ = '阿布'
__weixin__ = 'abu_quant'

__all__ = [
           'run_silhouette_cv_estimator',
           'run_prob_cv_estimator',
           'run_cv_estimator',
           'plot_learning_curve',
           'plot_decision_boundary',
           'plot_confusion_matrices',
           'plot_roc_estimator',
           'graphviz_tree',
           'visualize_tree'
           ]


# noinspection PyUnresolvedReferences
def run_silhouette_cv_estimator(estimator, x, n_folds=10):
    """
    只针对kmean的cv验证，使用silhouette_score对聚类后的结果labels_
    进行度量使用silhouette_score，kmean的cv验证只是简单的通过np.random.choice
    进行随机筛选x数据进行聚类的silhouette_score度量，并不涉及训练集测试集
    :param estimator: keman或者支持estimator.labels_, 只通过if not isinstance(estimator, ClusterMixin)进行过滤
    :param x: x特征矩阵
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: eg: array([ 0.693 ,  0.652 ,  0.6845,  0.6696,  0.6732,  0.6874,  0.668 ,
                         0.6743,  0.6748,  0.671 ])
    """

    if not isinstance(estimator, ClusterMixin):
        print('estimator must be ClusterMixin')
        return

    silhouette_list = list()
    # eg: n_folds = 10, len(x) = 150 -> 150 * 0.9 = 135
    choice_cnt = int(len(x) * ((n_folds - 1) / n_folds))
    choice_source = np.arange(0, x.shape[0])

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    for _ in np.arange(0, n_folds):
        # 只是简单的通过np.random.choice进行随机筛选x数据
        choice_index = np.random.choice(choice_source, choice_cnt)
        x_choice = x[choice_index]
        estimator.fit(x_choice)
        # 进行聚类的silhouette_score度量
        silhouette_score = metrics.silhouette_score(x_choice, estimator.labels_, metric='euclidean')
        silhouette_list.append(silhouette_score)
    return silhouette_list


def run_prob_cv_estimator(estimator, x, y, n_folds=10):
    """
    通过KFold和参数n_folds拆分训练集和测试集，使用
    np.zeros((len(y), len(np.unique(y))))初始化prob矩阵，
    通过训练estimator.fit(x_train, y_train)后的分类器使用
    predict_proba将y_prob中的对应填数据

    :param estimator: 支持predict_proba的有监督学习, 只通过hasattr(estimator, 'predict_proba')进行过滤
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: eg: y_prob
                array([[ 0.8726,  0.1274],
                       [ 0.0925,  0.9075],
                       [ 0.2485,  0.7515],
                       ...,
                       [ 0.3881,  0.6119],
                       [ 0.7472,  0.2528],
                       [ 0.8555,  0.1445]])

    """
    if not hasattr(estimator, 'predict_proba'):
        print('estimator must has predict_proba')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    kf = KFold(len(y), n_folds=n_folds, shuffle=True)
    y_prob = np.zeros((len(y), len(np.unique(y))))
    """
        根据y序列的数量以及y的label数量构造全是0的矩阵
        eg: y_prob
        array([[ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
                ..............
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
    """

    for train_index, test_index in kf:
        x_train, x_test = x[train_index], x[test_index]
        y_train = y[train_index]

        # clf = clone(estimator)
        estimator.fit(x_train, y_train)
        # 使用predict_proba将y_prob中的对应填数据
        y_prob[test_index] = estimator.predict_proba(x_test)

    return y_prob


def run_cv_estimator(estimator, x, y, n_folds=10):
    """
    通过KFold和参数n_folds拆分训练集和测试集，使用
    y.copy()初始化y_pred矩阵，迭代切割好的训练集与测试集，
    不断通过 estimator.predict(x_test)将y_pred中的值逐步替换

    :param estimator: 有监督学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: y_pred序列
    """
    if not hasattr(estimator, 'predict'):
        print('estimator must has predict')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    kf = KFold(len(y), n_folds=n_folds, shuffle=True)
    # 首先copy一个一摸一样的y
    y_pred = y.copy()
    """
        eg: y_pred
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])
    """

    for train_index, test_index in kf:
        x_train, x_test = x[train_index], x[test_index]
        y_train = y[train_index]
        estimator.fit(x_train, y_train)
        # 通过 estimator.predict(x_test)将y_pred中的值逐步替换
        y_pred[test_index] = estimator.predict(x_test)
    return y_pred


# warnings_filter针对多标签使用OneVsRestClassifier出现的版本警告
@warnings_filter
def plot_learning_curve(estimator, x, y, cv=5, n_jobs=1,
                        train_sizes=np.linspace(.05, 1., 20)):
    """
    绘制学习曲线，train_sizes使用np.linspace(.05, 1., 20)即训练集从5%－100%递进

            np.linspace(.05, 1., 20)
            array([ 0.05,  0.1 ,  0.15,  0.2 ,  0.25,  0.3 ,  0.35,  0.4 ,  0.45,
                    0.5 ,  0.55,  0.6 ,  0.65,  0.7 ,  0.75,  0.8 ,  0.85,  0.9 ,
                    0.95,  1.  ])

    套接sklern中learning_curve函数，传递estimator，cv等参数

    :param estimator: 学习器对象，透传learning_curve
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param cv: 透传learning_curve，cv参数，默认5，int
    :param n_jobs: 透传learning_curve，并行进程数，默认1，即使用单进程执行
    :param train_sizes: train_sizes使用np.linspace(.05, 1., 20)即训练集从5%－100%递进
    """

    # 套接learning_curve，返回训练集和测试集的score和对应的size
    train_sizes, train_scores, test_scores = learning_curve(
        estimator, x, y, cv=cv, n_jobs=n_jobs, train_sizes=train_sizes)
    """
        eg:  train_scores shape = (20, 5)
        array([[ 0.8571,  0.9143,  0.9143,  0.9143,  0.9143],
               [ 0.8169,  0.8732,  0.8732,  0.8732,  0.8732],
               [ 0.8208,  0.8396,  0.8396,  0.8396,  0.8396],
               [ 0.8028,  0.8099,  0.8099,  0.8099,  0.8099],
               [ 0.8146,  0.8202,  0.8146,  0.8146,  0.8146],
               [ 0.8263,  0.8263,  0.8216,  0.8216,  0.8216],
               [ 0.8153,  0.8273,  0.8112,  0.8112,  0.8112],
               [ 0.8063,  0.8169,  0.7993,  0.7993,  0.7993],
               [ 0.8156,  0.8281,  0.8063,  0.8063,  0.8063],
               [ 0.8169,  0.8254,  0.8254,  0.8254,  0.8254],
               [ 0.8184,  0.8235,  0.8261,  0.8312,  0.8312],
               [ 0.815 ,  0.822 ,  0.8197,  0.822 ,  0.822 ],
               [ 0.816 ,  0.8203,  0.8203,  0.8182,  0.8182],
               [ 0.8133,  0.8173,  0.8173,  0.8253,  0.8253],
               [ 0.8109,  0.8127,  0.8146,  0.8202,  0.8221],
               [ 0.8155,  0.819 ,  0.8172,  0.8207,  0.8225],
               [ 0.8149,  0.8248,  0.8231,  0.8248,  0.8198],
               [ 0.8187,  0.8281,  0.825 ,  0.8328,  0.8219],
               [ 0.8254,  0.8299,  0.8284,  0.8343,  0.8166],
               [ 0.8272,  0.8315,  0.8301,  0.8343,  0.8174]])
    """
    train_scores_mean = np.mean(train_scores, axis=1)
    """
        eg: train_scores_mean
            array([ 0.9029,  0.862 ,  0.8358,  0.8085,  0.8157,  0.8235,  0.8153,
            0.8042,  0.8125,  0.8237,  0.8261,  0.8201,  0.8186,  0.8197,
            0.8161,  0.819 ,  0.8215,  0.8253,  0.8269,  0.8281])
    """
    train_scores_std = np.std(train_scores, axis=1)

    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    """
         eg: test_scores_std
            array([ 0.0751,  0.0607,  0.0314,  0.0059,  0.0047,  0.0066,  0.0074,
                    0.0051,  0.0107,  0.0115,  0.0107,  0.012 ,  0.0142,  0.018 ,
                    0.0134,  0.0167,  0.0167,  0.0127,  0.0128,  0.0113])
    """
    # 开始可视化学习曲线
    plt.figure()
    plt.title('learning curve')
    plt.xlabel("train sizes")
    plt.ylabel("scores")
    plt.gca().invert_yaxis()
    plt.grid()
    # 对train_scores的均值和方差区域进行填充
    plt.fill_between(train_sizes, train_scores_mean - train_scores_std, train_scores_mean + train_scores_std,
                     alpha=0.1, color="g")
    # 对test_scores的均值和方差区域进行填充
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std, test_scores_mean + test_scores_std,
                     alpha=0.1, color="r")
    # 把train_scores_mean标注圆圈
    plt.plot(train_sizes, train_scores_mean, 'o-', color="g", label="train scores")
    # 把ttest_scores_mean标注圆圈
    plt.plot(train_sizes, test_scores_mean, 'o-', color="r", label="test scores")
    plt.legend(loc="best")
    plt.draw()
    plt.gca().invert_yaxis()
    plt.show()


def graphviz_tree(estimator, features, x, y):
    """
    绘制决策树或者core基于树的分类回归算法的决策示意图绘制，查看
    学习器本身hasattr(fiter, 'tree_')是否有tree_属性，内部clone(estimator)学习器
    后再进行训练操作，完成训练后使用sklearn中tree.export_graphvizd导出graphviz.dot文件
    需要使用第三方dot工具将graphviz.dot进行转换graphviz.png，即内部实行使用
    运行命令行：
                os.system("dot -T png graphviz.dot -o graphviz.png")
    最后读取决策示意图显示

    :param estimator: 学习器对象，透传learning_curve
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param features: 训练集x矩阵列特征所队员的名称，可迭代序列对象
    """
    if not hasattr(estimator, 'tree_'):
        logging.info('only tree can graphviz!')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    estimator.fit(x, y)
    # TODO out_file path放倒cache中
    tree.export_graphviz(estimator.tree_, out_file='graphviz.dot', feature_names=features)
    os.system("dot -T png graphviz.dot -o graphviz.png")

    '''
        !open $path
        要是方便用notebook直接open其实显示效果好，plt，show的大小不好调整
    '''
    graphviz = os.path.join(os.path.abspath('.'), 'graphviz.png')

    # path = graphviz
    # !open $path
    if not file_exist(graphviz):
        logging.info('{} not exist! please install dot util!'.format(graphviz))
        return

    image_file = cbook.get_sample_data(graphviz)
    image = plt.imread(image_file)
    image_file.close()
    plt.imshow(image)
    plt.axis('off')  # clear x- and y-axes
    plt.show()


def visualize_tree(estimator, x, y, boundaries=True):
    """
    需要x矩阵特征列只有两个维度，根据x，y，通过meshgrid构造训练集平面特征
    通过z = estimator.predict(np.c_[xx.ravel(), yy.ravel()])对特征平面
    进行predict生成z轴，可视化meshgrid构造训练集平面特征使用生成的z生成
    pcolormesh进行可视化

    :param estimator: 学习器对象，内部clone(estimator)
    :param x: 训练集x矩阵，numpy矩阵，需要特征列只有两个维度
    :param y: 训练集y序列，numpy序列
    :param boundaries: 是否绘制决策边界

    """
    if x.shape[1] != 2:
        logging.info('be sure x shape[1] == 2!')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    estimator.fit(x, y)

    xlim = (x[:, 0].min() - 0.1, x[:, 0].max() + 0.1)
    ylim = (x[:, 1].min() - 0.1, x[:, 1].max() + 0.1)
    x_min, x_max = xlim
    y_min, y_max = ylim
    # 通过训练集中x的min和max，y的min，max构成meshgrid
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 100),
                         np.linspace(y_min, y_max, 100))
    # 摊平xx，yy进行z轴的predict
    z = estimator.predict(np.c_[xx.ravel(), yy.ravel()])
    # z的shape跟随xx
    z = z.reshape(xx.shape)
    plt.figure()
    plt.pcolormesh(xx, yy, z, alpha=0.2, cmap='rainbow')
    plt.clim(y.min(), y.max())

    # 将之前的训练集中的两个特征进行scatter绘制，颜色使用y做区分
    plt.scatter(x[:, 0], x[:, 1], c=y, s=50, cmap='rainbow')
    plt.axis('off')

    plt.xlim(x_min, x_max)
    plt.ylim(y_min, y_max)
    plt.clim(y.min(), y.max())

    def plot_boundaries(i, p_xlim, p_ylim):
        """
        针对有tree_属性的学习器绘制决策边界
        :param i: 内部递归调用使用ree_inner.children_left[i]和tree_inner.children_right[i]
        :param p_xlim: 原始参数使用plt.xlim()
        :param p_ylim: 原始参数使用plt.ylim()
        """
        if i < 0:
            return
        # 拿到tree_使用plot_boundaries继续递归绘制
        tree_inner = estimator.tree_

        if tree_inner.feature[i] == 0:
            # 绘制0的边界
            plt.plot([tree_inner.threshold[i], tree_inner.threshold[i]], p_ylim, '-k')
            # 即x轴固定p_ylim，xlim=[p_xlim[0], tree_inner.threshold[i]], [tree_inner.threshold[i], p_xlim[1]]
            plot_boundaries(tree_inner.children_left[i],
                            [p_xlim[0], tree_inner.threshold[i]], p_ylim)
            plot_boundaries(tree_inner.children_right[i],
                            [tree_inner.threshold[i], p_xlim[1]], p_ylim)
        elif tree_inner.feature[i] == 1:
            # 绘制1的边界
            plt.plot(p_xlim, [tree_inner.threshold[i], tree_inner.threshold[i]], '-k')
            # 即y轴固定p_xlim，ylim=[p_ylim[0], tree_inner.threshold[i]], [tree_inner.threshold[i], p_ylim[1]]
            plot_boundaries(tree_inner.children_left[i], p_xlim,
                            [p_ylim[0], tree_inner.threshold[i]])
            plot_boundaries(tree_inner.children_right[i], p_xlim,
                            [tree_inner.threshold[i], p_ylim[1]])

    if boundaries and hasattr(estimator, 'tree_'):
        # 简单决策树才去画决策边界
        plot_boundaries(0, plt.xlim(), plt.ylim())


def plot_decision_boundary(pred_func, x, y):
    """
    通过x，y以构建meshgrid平面区域，要x矩阵特征列只有两个维度，在区域中使用外部传递的
    pred_func函数进行z轴的predict，通过contourf绘制特征平面区域，最后使用
    plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.Spectral)在平面区域上填充原始特征
    点

    :param pred_func: callable函数，eg：pred_func: lambda p_x: fiter.predict(p_x), x, y
    :param x: 训练集x矩阵，numpy矩阵，需要特征列只有两个维度
    :param y: 训练集y序列，numpy序列
    """
    xlim = (x[:, 0].min() - 0.1, x[:, 0].max() + 0.1)
    ylim = (x[:, 1].min() - 0.1, x[:, 1].max() + 0.1)
    x_min, x_max = xlim
    y_min, y_max = ylim
    # 通过训练集中x的min和max，y的min，max构成meshgrid
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 200),
                         np.linspace(y_min, y_max, 200))
    # 摊平xx，yy进行z轴的predict, pred_func: lambda p_x: fiter.predict(p_x), x, y
    z = pred_func(np.c_[xx.ravel(), yy.ravel()])
    # z的shape跟随xx
    z = z.reshape(xx.shape)
    # 使用contourf绘制xx, yy, z，即特征平面区域以及z的颜色区别
    # noinspection PyUnresolvedReferences
    plt.contourf(xx, yy, z, cmap=plt.cm.Spectral)
    # noinspection PyUnresolvedReferences
    # 在特征区域的基础上将原始，两个维度使用scatter绘制以y为颜色的点
    plt.scatter(x[:, 0], x[:, 1], c=y, cmap=plt.cm.Spectral)
    plt.show()


def plot_roc_estimator(estimator, x, y, pos_label=None):
    """
    固定n_folds=10通过kf = KFold(len(y), n_folds=10, shuffle=True)拆分
    训练测试集，使用estimator.predict_proba对测试集数据进行概率统计，直接使用
    sklearn中的roc_curve分别对多组测试集计算fpr, tpr, thresholds，并计算roc_auc
    最后绘制roc_auc曲线进行可视化操作

    :param estimator: 分类器对象，内部clone(estimator)
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param pos_label: 对y大于2个label的数据，roc_curve需要指定pos_label，如果不指定，默认使用y的第一个label值
    """

    if not hasattr(estimator, 'predict_proba'):
        # 分类器必须要有predict_proba方法
        logging.info('estimator must has predict_proba!')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    estimator.fit(x, y)

    # eg: y_unique = [0, 1]
    y_unique = np.unique(y)
    kf = KFold(len(y), n_folds=10, shuffle=True)
    y_prob = np.zeros((len(y), len(y_unique)))
    """
        eg:  y_prob
            array([[ 0.,  0.],
                   [ 0.,  0.],
                   [ 0.,  0.],
                   ...,
                   [ 0.,  0.],
                   [ 0.,  0.],
                   [ 0.,  0.]])
    """
    mean_tpr = 0.0
    # 0-1分布100个
    mean_fpr = np.linspace(0, 1, 100)
    for i, (train_index, test_index) in enumerate(kf):
        x_train, x_test = x[train_index], x[test_index]
        y_train = y[train_index]
        estimator.fit(x_train, y_train)
        y_prob[test_index] = estimator.predict_proba(x_test)
        """
            eg: y_prob[test_index]
            array([[ 0.8358,  0.1642],
                   [ 0.4442,  0.5558],
                   [ 0.1351,  0.8649],
                   [ 0.8567,  0.1433],
                   [ 0.6953,  0.3047],
                   ..................
                   [ 0.1877,  0.8123],
                   [ 0.8465,  0.1535],
                   [ 0.1916,  0.8084],
                   [ 0.8421,  0.1579]])

        """
        if len(y_unique) != 2 and pos_label is None:
            # 对y大于2个label的数据，roc_curve需要指定pos_label，如果不指定，默认使用y的第一个label值
            pos_label = y_unique[0]
            logging.info('y label count > 2 and param pos_label is None, so choice y_unique[0]={} for pos_label!'.
                         format(pos_label))

        fpr, tpr, thresholds = roc_curve(y[test_index], y_prob[test_index, 1], pos_label=pos_label)
        """
            eg:
                fpr
                array([ 0.    ,  0.0169,  0.0169,  0.0339,  0.0339,  0.0508,  0.0508,
                        0.0847,  0.0847,  0.1017,  0.1017,  0.1186,  0.1186,  0.2034,
                        0.2034,  0.2542,  0.2542,  0.5254,  0.5254,  0.5763,  0.5763,  1.    ])
                tpr
                array([ 0.0323,  0.0323,  0.4839,  0.4839,  0.5484,  0.5484,  0.6452,
                        0.6452,  0.7419,  0.7419,  0.7742,  0.7742,  0.8387,  0.8387,
                        0.9032,  0.9032,  0.9355,  0.9355,  0.9677,  0.9677,  1.    ,  1.    ])

                thresholds
                array([ 0.9442,  0.9288,  0.8266,  0.8257,  0.8123,  0.8122,  0.8032,
                        0.7647,  0.7039,  0.5696,  0.5558,  0.4854,  0.4538,  0.2632,
                        0.2153,  0.2012,  0.1902,  0.1616,  0.1605,  0.1579,  0.1561,
                        0.1301])
        """

        # interp线性插值计算
        mean_tpr += interp(mean_fpr, fpr, tpr)
        # 把第一个值固定0，最后会使用mean_tpr[-1] = 1.0把最后一个固定1.0
        mean_tpr[0] = 0.0
        # 直接使用 sklearn中的metrics.auc计算
        roc_auc = auc(fpr, tpr)
        plt.plot(fpr, tpr, lw=1, label='ROC fold %d (area = %0.2f)' % (i, roc_auc))

    mean_tpr /= len(kf)
    # 最后一个固定1.0
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    plt.plot(mean_fpr, mean_tpr, 'k--', label='Mean ROC (area = %0.2f)' % mean_auc, lw=2)

    plt.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6), label='Random')
    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic')
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.show()


def plot_confusion_matrices(estimator, x, y, n_folds=10):
    """
    套接run_cv_estimator进行通过参数n_folds进行训练集测试集拆封
    使用y_pred和y做为参数，透传给metrics.confusion_matrix函数
    进行混淆矩阵的计算，通过ax.matshow可视化混淆矩阵

    :param estimator: 分类器对象，内部clone(estimator)
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_folds: 透传KFold参数，切割训练集测试集参数
    """
    y_pred = run_cv_estimator(estimator, x, y, n_folds=n_folds)
    """
        eg: y_pred
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
           1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1,
           1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 2, 2,
           2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
           2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])
    """
    y_unique = np.unique(y)

    confusion_matrix = metrics.confusion_matrix(y, y_pred)
    """
        eg: confusion_matrix
         [[50  0  0]
         [ 0 47  3]
         [ 0  1 49]]
    """
    logging.info(confusion_matrix)

    fig = plt.figure()
    # 颜色条的颜色数量设置使用len(y_unique) * len(y_unique)，即如果y是3个label->9颜色。2->4
    cmap = plt.get_cmap('jet', len(y_unique) * len(y_unique))
    cmap.set_under('gray')
    ax = fig.add_subplot(111)
    # ax.matshow可视化化混淆矩阵
    cax = ax.matshow(confusion_matrix, cmap=cmap,
                     vmin=confusion_matrix.min(),
                     vmax=confusion_matrix.max())
    plt.title('Confusion matrix for %s' % estimator.__class__.__name__)
    # 辅助颜色边bar显示
    fig.colorbar(cax)
    # noinspection PyTypeChecker
    ax.set_xticklabels('x: '.format(y_unique))
    # noinspection PyTypeChecker
    ax.set_yticklabels('y: '.format(y_unique))
    plt.xlabel('Predicted')
    plt.ylabel('True')
    plt.show()



# -*- encoding:utf-8 -*-
"""量化技术分析工具图形可视化基础模块"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging
import functools
from contextlib import contextmanager

import ipywidgets as widgets
import pandas as pd

from ..CoreBu import ABuEnv
from ..WidgetBu.ABuWGBase import WidgetBase, show_msg_func, browser_down_csv_zip
from ..WidgetBu.ABuWGBSymbol import WidgetSymbolChoice
from ..WidgetBu.ABuWGBRunBase import WidgetEnvSetMixin, WidgetTimeModeMixin
from ..MarketBu import ABuSymbolPd
from ..UtilBu import ABuProgress
from ..TLineBu.ABuTL import AbuTLine

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetToolSet(WidgetBase, WidgetEnvSetMixin, WidgetTimeModeMixin):
    """基础设置界面"""

    def __init__(self):
        """初始化基础设置界面"""

        mdm_box = self.init_env_set_ui()
        tm_box = self.init_time_mode_ui()
        self.sc = WidgetSymbolChoice()
        self.sc.cs_tip.value = u'如果股池为空，将使用示例的symbol进行分析'
        # 默认1年的数据分析
        self.run_years.value = 1
        # self.widget = widgets.Box([self.sc.widget, tm_box, mdm_box], layout=self.widget_layout)
        self.widget = widgets.VBox([self.sc.widget, tm_box, mdm_box])

    def time_mode_str(self):
        """实现混入WidgetTimeModeMixin，声明时间模块代表分析"""
        return u'分析'


def single_fetch_symbol_analyse(func):
    """定参数装饰器函数：获取设置中的1个symbol进行数据获取后调用被装饰函数"""

    @functools.wraps(func)
    def wrapper(self, bt):
        symbol = self._choice_symbol_single()
        kl = self._fetch_single_kl(symbol)

        ABuProgress.clear_output()
        if kl is not None:
            # 清理之前的输出结果
            kl_tl = AbuTLine(kl.close, 'kl')
            return func(self, kl, kl_tl, bt)
        else:
            self.info_change_set_mode(symbol)

    return wrapper


def multi_fetch_symbol_analyse(func):
    """定参数装饰器函数：获取设置中的多个symbol数据获取组成字典后调用被装饰函数"""

    @functools.wraps(func)
    def wrapper(self, bt):
        choice_symbol = self._choice_symbol_multi()
        kl_dict = self._fetch_multi_kl(choice_symbol)
        ABuProgress.clear_output()
        if kl_dict is not None and len(kl_dict) > 0:
            return func(self, kl_dict, bt)
        else:
            self.info_change_set_mode(choice_symbol)

    return wrapper


def multi_fetch_symbol_df_analyse(col_key):
    """
    定参数装饰器函数：获取设置中的多个symbol数据获取组成字典后
    通过参数col_key获取所有金融序列中的某一列形成新的dataframe，
    将dataframe传递给被装饰函数
    :param col_key: 金融序列中的某一列名称，eg：p_change, close
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(self, bt):
            choice_symbol = self._choice_symbol_multi()
            cg_df = self._fetch_multi_kl_col(choice_symbol, col_key=col_key)
            ABuProgress.clear_output()
            if cg_df is not None and len(cg_df) > 0:
                return func(self, cg_df, bt)
            else:
                self.info_change_set_mode(choice_symbol)

        return wrapper

    return decorate


# noinspection PyUnusedLocal,PyProtectedMember
class WidgetToolBase(WidgetBase):
    """技术分析基类"""

    def __init__(self, tool_set):
        """
            构造通用的layout css:
            label_layout, tool_layout, scroll_widget_layout,
            description_layout
        """
        if not isinstance(tool_set, WidgetToolSet):
            raise TypeError('tool_set must isinstance WidgetToolSet, not {}!'.format(type(tool_set)))
        self.tool_set = tool_set
        self.label_layout = widgets.Layout(width='300px', align_items='stretch')
        self.tool_layout = widgets.Layout(align_items='stretch', justify_content='space-between')
        self.scroll_widget_layout = widgets.Layout(overflow_x='scroll',
                                                   flex_direction='row',
                                                   display='flex')
        self.description_layout = widgets.Layout(height='150px')
        # 默认不启动可滚动因子界面，因为对外的widget版本以及os操作系统不统一
        self.scroll_factor_box = False

    def _sub_children(self, children, n_split):
        """将children每n_split个为一组，组装子children_group序列"""
        sub_children_cnt = int(len(children) / n_split)
        if sub_children_cnt == 0:
            sub_children_cnt = 1
        group_adjacent = lambda a, k: zip(*([iter(a)] * k))
        children_group = list(group_adjacent(children, sub_children_cnt))
        residue_ind = -(len(children) % sub_children_cnt) if sub_children_cnt > 0 else 0
        if residue_ind < 0:
            children_group.append(children[residue_ind:])
        return children_group

    @contextmanager
    def data_mode_recover(self, is_example_mode):
        """
        上下文管理器函数：
        1. 上文通过参数is_example_mode与全局的沙盒模式是否相同改变设置
        2. 下文承接是否上文修改了沙盒模式设置，恢复上文的修改
        :param is_example_mode: 临时设置中是否使用沙盒数据模式
        """
        recover = False
        if is_example_mode != ABuEnv._g_enable_example_env_ipython:
            # 不要使用enable_example_env_ipython改变缓存模式，只临时改变沙盒类型
            ABuEnv._g_enable_example_env_ipython = is_example_mode
            recover = True

        yield
        if recover:
            # 不要使用disable_example_env_ipython改变缓存模式，只恢复沙盒类型
            ABuEnv._g_enable_example_env_ipython = not is_example_mode

    def map_tip_target_label(self, n_target):
        """
        根据n_target要在设置中选择几个symbol映射文字提示
        :param n_target: eg：1，2，－1
        :return: eg：需选择1个分析目标在分析设置中
        """
        if n_target == 1:
            return u'需选择1个分析目标在\'分析设置\'中'
        elif n_target == 2:
            return u'需选择2个分析目标在\'分析设置\'中'
        else:
            return u'需选择多个(>1个)分析目标在\'分析设置\'中'

    def _start_end_n_fold(self):
        """获取设置中的时间获取模式以及具体的start，end，n_folds"""
        n_folds = 1
        start = None
        end = None
        if not self.tool_set.run_years.disabled:
            # 如果使用年回测模式
            n_folds = self.tool_set.run_years.value
        if not self.tool_set.start.disabled:
            # 使用开始回测日期
            start = self.tool_set.start.value
        if not self.tool_set.end.disabled:
            # 使用结束回测日期
            end = self.tool_set.end.value
        return start, end, n_folds

    def info_change_set_mode(self, symbol):
        """
            在上层定参数装饰器single_fetch_symbol_analyse，multi_fetch_symbol_analyse
            中统一处理数据获取失败的case，如果沙盒模式提醒改变设置数据更新
        """
        # noinspection PyProtectedMember
        if ABuEnv._g_enable_example_env_ipython:
            logging.info(
                u'当前数据模式为\'沙盒模式\'无{}数据，'
                u'请在\'分析设置\'中切换数据模式并确认数据可获取！'
                u'非沙盒模式建议先用\'数据下载界面操作\'进行数据下载'
                u'之后设置数据模式为\'开放数据模式\'，联网模式使用\'本地数据模式\''.format(symbol))
        else:
            logging.info(u'{}数据获取失败！'.format(symbol))
        browser_down_csv_zip()

    def _fetch_single_kl(self, symbol):
        """
        通过_start_end_n_fold获取时间参数进行金融序列获取
        :param symbol: eg: usTSLA
        :return: pd.DataFrame
        """
        start, end, n_folds = self._start_end_n_fold()
        kl = ABuSymbolPd.make_kl_df(symbol, n_folds=n_folds, start=start, end=end)
        return kl

    def _fetch_multi_kl(self, choice_symbol):
        """
        多个symbol目标，通过_start_end_n_fold获取时间参数进行金融序列获取
        :param choice_symbol: eg: ['usTSLA', 'usNOAH']
        :return: eg: {'usTSLA': pd.DataFrame, 'usNOAH': pd.DataFrame}
        """
        start, end, n_folds = self._start_end_n_fold()

        kl_dict = {symbol: ABuSymbolPd.make_kl_df(symbol, start=start, end=end, n_folds=n_folds)
                   for symbol in choice_symbol}
        kl_dict = {kl_key: kl_dict[kl_key] for kl_key in kl_dict if kl_dict[kl_key] is not None}
        return kl_dict

    def _fetch_multi_kl_col(self, choice_symbol, col_key, na_val=0):
        """
         多个symbol目标，通过_start_end_n_fold获取时间参数进行金融序列获取,
         多个symbol组成的字典对象通过col_key获取各个元素的金融序列中对应的列
         形成新的pd.DateFrame对象
         :param choice_symbol: eg: ['usTSLA', 'usNOAH']
         :return: pd.DateFrame对象
         """
        start, end, n_folds = self._start_end_n_fold()

        kl_dict = {symbol: ABuSymbolPd.make_kl_df(symbol, start=start, end=end, n_folds=n_folds)
                   for symbol in choice_symbol}
        kl_dict = {kl_key: kl_dict[kl_key] for kl_key in kl_dict if kl_dict[kl_key] is not None}

        if len(kl_dict) > 0:
            kl_col_df = pd.concat({kl_name: kl_dict[kl_name][col_key] for kl_name in kl_dict}, axis=1)
            # noinspection PyUnresolvedReferences
            kl_col_df = kl_col_df.fillna(value=na_val)
            return kl_col_df

    def _choice_symbol_single(self, default=None):
        """单独一个分析目标的函数"""
        choice_symbols = self.tool_set.sc.choice_symbols.options
        if choice_symbols is None or len(choice_symbols) == 0:
            # 如果一个symbol都没有，使用示例
            symbol = 'usTSLA' if default is None else default
            show_msg_func(u'未设置任何symbol将使用示例{}进行分析'.format(symbol))
        elif choice_symbols is not None and len(choice_symbols) > 1:
            symbol = choice_symbols[0]
            show_msg_func(u'分析设置多个symbol目标，只取第一个{}进行分析'.format(symbol))
        else:
            symbol = choice_symbols[-1]
        return symbol

    def _choice_symbol_pair(self, default=None):
        """两个分析目标的函数"""
        choice_symbols = self.tool_set.sc.choice_symbols.options
        if choice_symbols is None or len(choice_symbols) < 2:
            # 如果一个symbol都没有，使用示例
            if default is None:
                symbol1, symbol2 = 'AU0', 'XAU'
            else:
                symbol1, symbol2 = default[0], default[1]
            show_msg_func(u'需要选择两个symbol，将使用示例{} vs {}进行分析'.format(symbol1, symbol2))
        elif choice_symbols is not None and len(choice_symbols) > 2:
            symbol1, symbol2 = choice_symbols[0], choice_symbols[1]
            show_msg_func(u'分析设置多个symbol目标，只取第一个{}，第二个{}进行分析'.format(symbol1, symbol2))
        else:
            symbol1, symbol2 = choice_symbols[-2], choice_symbols[-1]
        return symbol1, symbol2

    def _choice_symbol_multi(self):
        """分析多个目标的函数"""
        choice_symbols = self.tool_set.sc.choice_symbols.options
        if choice_symbols is None or len(choice_symbols) == 0:
            # 如果一个symbol都没有，使用示例
            choice_symbols = ['sh600036', 'sh600809', 'hk00700', 'hk03333']
            show_msg_func(u'未设置任何symbol将使用示例')
        return choice_symbols



from __future__ import absolute_import

from .ABuBenchmark import AbuBenchmark
from .ABuCapital import AbuCapital
from .ABuKLManager import AbuKLManager
from .ABuOrder import AbuOrder

from . import ABuMLFeature as feature
from .ABuMLFeature import AbuFeatureDegExtend

from .ABuMLFeature import AbuFeatureBase, BuyFeatureMixin, SellFeatureMixin
from .ABuTradeProxy import AbuOrderPdProxy, EOrderSameRule


from . import ABuTradeDrawer
from . import ABuTradeExecute
from . import ABuTradeProxy

__all__ = [
    'AbuBenchmark',
    'AbuCapital',
    'AbuKLManager',
    'AbuOrder',
    'AbuOrderPdProxy',
    'EOrderSameRule',

    'feature',
    'AbuFeatureDegExtend',
    'AbuFeatureBase',
    'BuyFeatureMixin',
    'SellFeatureMixin',
    'ABuTradeDrawer',
    'ABuTradeExecute',
    'ABuTradeProxy']



from __future__ import absolute_import

# noinspection PyUnresolvedReferences
from .ABuMLBinsCs import *
# noinspection PyUnresolvedReferences
from .ABuMLExecute import *
# noinspection PyUnresolvedReferences
from .ABuMLGrid import *
# noinspection PyUnresolvedReferences
from .ABuML import EMLFitType
# noinspection PyUnresolvedReferences
from .ABuMLPd import BtcBigWaveClf, ClosePredict



# coding=utf-8

import os
import sys
from os import path
import logging
import shutil
from setuptools import setup, find_packages

import abupy

DIST_NAME = 'abupy'
LICENSE = 'GPL'
AUTHOR = u"阿布"
EMAIL = "service@abuquant.com"
URL = "http://abuquant.com/"
DOWNLOAD_URL = 'https://github.com/bbfamily/abu'
CLASSIFIERS = []

DESCRIPTION = "阿布量化系统"
LONG_DESCRIPTION = """
**abu追求的是一句话就能够说明的智能策略**
abu能够帮助用户自动完善策略，主动分析策略产生的交易行为，智能拦截策略生成的容易失败的交易单。

现阶段的量化策略还是人工编写的代码，abu量化交易系统的设计将会向着由计算机自动实现整套流程的方向迈进，包括编写量化策略本身。

我们对未来的期望是：abupy用户只需要提供一些简单的种子策略，计算机在这些种子基础上不断自我学习、自我成长，创造出新的策略，并且随着时间序列数据不断智能调整策略的参数。

### 特点

* 使用多种机器学习技术智能优化策略
* 在实盘中指导策略进行交易，提高策略的实盘效果，战胜市场

### 支持的投资市场:

* 美股，A股，港股
* 期货，期权
* 比特币，莱特币

### 工程设计目标：

* 分离基础策略和策略优化监督模块
* 提高灵活度和适配性

"""
KEY_WORDS = ['阿布', 'abu', 'quant', 'quantization',
             'crawler', 'spider', 'scrapy', 'stock', 'machine learning',
             '股票', '机器学习', '量化', '爬虫']


def init_mete_data():
    mete_data = dict()
    mete_data['name'] = DIST_NAME
    mete_data['version'] = abupy.__version__
    mete_data['url'] = URL
    mete_data['download_url'] = DOWNLOAD_URL
    mete_data['description'] = DESCRIPTION
    mete_data['long_description'] = LONG_DESCRIPTION
    mete_data['author'] = AUTHOR
    mete_data['license'] = LICENSE
    mete_data['author_email'] = EMAIL
    mete_data['platforms'] = 'any'
    mete_data['keywords'] = KEY_WORDS

    mete_data['packages'] = list(filter(lambda pack: pack.startswith(DIST_NAME), find_packages()))
    mete_data['package_data'] = {'abupy': ['RomDataBu/*.txt',
                                           'RomDataBu/*.db',
                                           'RomDataBu/*.csv',
                                           'RomDataBu/*.zip',
                                           'RomDataBu/us_industries']}

    mete_data['install_requires'] = ['numpy',
                                     'pandas',
                                     'scipy',
                                     'scikit-learn',
                                     'matplotlib',
                                     'seaborn',
                                     'statsmodels',
                                     'requests']

    return mete_data


def clear_build():
    from abupy.UtilBu.ABuFileUtil import file_exist
    cur_path = path.dirname(path.abspath(path.realpath(__file__)))
    if file_exist(path.join(cur_path, 'dist')):
        shutil.rmtree('dist')
    if file_exist(path.join(cur_path, 'build')):
        shutil.rmtree('build')
    if file_exist(path.join(cur_path, 'abupy.egg-info')):
        shutil.rmtree('abupy.egg-info')


def uninstall():
    pass


def install():
    uninstall()
    clear_build()
    sys.argv.append('install')
    setup(**init_mete_data())


def deploy_2_pypi():
    """
    cmd: python setup.py sdist bdist_wheel && twine upload dist/*
    """
    clear_build()
    # 源码方式
    # e.g:  dist/abupy-xxxx-.tar.gz
    sys.argv.append('sdist')
    # python wheel 方式
    sys.argv.append('bdist_wheel')
    # wheel方式 通用包，即生成py2.py3的个平台包
    # e.g: dist/abupy-xxx-py2.py3-none-any.whl
    sys.argv.append('--universal')
    setup(**init_mete_data())
    """
    upload 需要在当前用户目录创建.pypirc文件，并写入
    ###########################################
    [distutils]
    index-servers=pypi

    [pypi]
    repository = https://upload.pypi.org/legacy/
    username = 你的用户名
    password = 你的密码
    ###########################################
    """

    def deploy():
        from twine.commands import upload
        upload.main(['dist/*'])

    try:
        deploy()
    except ImportError:
        if os.system('pip install twine') != 0:
            deploy()
        else:
            logging.warning('install twine error!! must be need root permission.')


if __name__ == '__main__':
    if '2pypi' in sys.argv:
        sys.argv.remove('2pypi')
        deploy_2_pypi()
    else:
        install()



# -*- encoding:utf-8 -*-
"""
    深度学习工具模块，为caffe等工具库标准化图片格式
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import glob
import imghdr
import os

import PIL.Image
from PIL import ImageFile

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import map

from ..UtilBu import ABuFileUtil

__all__ = ['std_img_from_root_dir',
           'covert_to_jpeg',
           'find_img_by_ext',
           'change_to_real_type'
           ]

__author__ = '阿布'
__weixin__ = 'abu_quant'

# 为了im.convert('RGB')的异常错误，需要设置ImageFile.LOAD_TRUNCATED_IMAGES = True
ImageFile.LOAD_TRUNCATED_IMAGES = True


def std_img_from_root_dir(root_dir, a_ext):
    """
    找到root_dir目录下所有.ext的图片，都转换其真实类型后，
    将不是jpeg的全部转换为jpeg，即标准化图像格式
    :param root_dir: str对象，文件夹路径信息
    :param a_ext: 图像文件后缀，eg. png，jpg
    """

    # root_dir目录下及一级子目录下所有.ext后缀的文件全路径返回
    img_list = find_img_by_ext(a_ext, root_dir)
    # img_list中所有的转换为其真实类型
    all_type = change_to_real_type(img_list)
    for ext in all_type:
        # 迭代change_to_real_type返回的集合
        if ext != 'jpeg':
            # 标准化所有输入图片，将不是jpeg的都转换为jpeg
            if ext is None:
                ext = a_ext
            # 找到非标准类型子序列sub_img_list
            sub_img_list = find_img_by_ext(ext, root_dir)
            # 一个个进行转换
            _ = list(map(lambda img: covert_to_jpeg(img), sub_img_list))
            # 然后再次转换其真实后缀名
            change_to_real_type(sub_img_list)


def covert_to_jpeg(org_img, dst_img=None):
    """
    将输入img转换为RGB的jpeg格式图像
    :param org_img: 原始图像路径，str对象
    :param dst_img: 转换后的图像路径输出路径，str对象，默认=None将覆盖org_img路径
    """
    im = PIL.Image.open(org_img)
    if dst_img is None:
        # 不设置输出路径，即在输入文件上直接覆盖
        dst_img = org_img
    im.convert('RGB').save(dst_img, 'JPEG')


def find_img_by_ext(ext, root_dir):
    """
    将root_dir目录下及一级子目录下所有.ext后缀的文件全路径返回,
    注意只遍历根目录及一级子目录
    :param ext: str对象 图像文件后缀，eg. png，jpg
    :param root_dir: str对象，文件夹路径信息
    :return: list序列对象
    """

    # 遍历一级子目录
    dirs = [root_dir + name for name in os.listdir(root_dir) if os.path.isdir(os.path.join(root_dir, name))]
    # 形成根目录及一级子目录
    dirs.append(root_dir)
    img_list = list()
    for dr in dirs:
        # 遍历所有目录，通过glob模糊查询所有.ext文件
        sub_list = glob.glob('{}/*.{}'.format(dr, ext))
        img_list.extend(sub_list)
    return img_list


def change_to_real_type(img_list):
    """
    将img的后缀名转换为其真实类型
        eg. a.png 如果 a.png实际上是jpeg，则将后缀修改－> a.jpeg
    :param img_list:
    :return:
    """
    # 类型记录集合
    record_type = set()
    for img in img_list:
        if not ABuFileUtil.file_exist(img):
            # 过滤实际不存在的文件
            continue

        # 使用imghdr识别图像真实类型
        real_type = imghdr.what(img)
        # 将img_list有的类型做记录，add到集合中
        record_type.add(real_type)
        if real_type is None:
            # 识别不出来，可能就是残图，过滤
            continue
        # 修改成真实的类型
        real_name = img[:img.rfind('.')] + '.' + real_type
        os.rename(img, real_name)
    return record_type



from __future__ import absolute_import

from .ABuML import AbuML
from .ABuMLCreater import AbuMLCreater
from .ABuMLPd import AbuMLPd
from . import ABuMLExecute
from . import ABuMLGrid

from . import ABuMLApi as ml

__all__ = [
    'AbuML',
    'AbuMLCreater',
    'AbuMLPd',
    'ABuMLExecute',
    'ABuMLGrid',

    'ml'
]



# -*- encoding:utf-8 -*-
"""股票基本信息图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging

import ipywidgets as widgets

from ..MarketBu.ABuSymbolStock import query_stock_info
from ..MarketBu import ABuIndustries
from ..MarketBu.ABuSymbol import code_to_symbol
from ..UtilBu.ABuDTUtil import catch_error
from ..CoreBu import ABuEnv
from ..CoreBu.ABuFixes import six
from ..UtilBu.ABuStrUtil import to_unicode
from ..WidgetBu.ABuWGBase import WidgetBase, WidgetSearchBox

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetSearchStockInfo(WidgetBase):
    def __init__(self):
        self.stock_info = None
        self.widget = WidgetSearchBox(self.on_search_result_click)()

    def on_search_result_click(self, select):
        """选中搜索结果中的一个进行点击"""

        st_symbol = [symbol.split(':')[1] if symbol.find(':') > 0
                     else symbol for symbol in list(select['new'])]
        if len(st_symbol) == 0:
            return

        # result是多选框，值拿最后一个选中的
        symbol = st_symbol[-1]
        if self.stock_info is not None and self.stock_info() is not None:
            self.stock_info().close()
        self.stock_info = WidgetStockInfo(symbol)
        self.stock_info.display()


class WidgetStockInfo(WidgetBase):
    """股票基本信息界面组件类"""

    def __init__(self, symbol_code):
        """
        构建股票基本信息
        :param symbol_code: 字符串symbol eg usTSLA 或者 Symbol对象
        """
        if isinstance(symbol_code, six.string_types):
            symbol_code = code_to_symbol(symbol_code)

        # 类中的symbol对象为Symbol类对象，即包含市场，子市场等信息的对象
        self.symbol = symbol_code
        # 默认规则所有组件最终ui成品为widget，暂时未使用基类来约束
        self.widget = self.stock_base_info(self.symbol)

    def other_industries_symbol(self, stock_info):
        """从股票信息查询其对应行业的其它股票，构建形势为按钮序列，暂时只取出10个相关的股票"""
        other_co_symbol = ABuIndustries.query_factorize_industry_symbol(
            stock_info.industry_factorize.values[0], market=self.symbol.market)
        other_co_bs = []
        # TODO 加载更多或者分段加载方式，暂时只取出10个相关的股票
        other_co_symbol = other_co_symbol[:10] if len(other_co_symbol) > 10 else other_co_symbol
        for symbol in other_co_symbol:
            # 通过symbol查询公司名称等信息
            stock_info = query_stock_info(symbol)
            if stock_info is None or stock_info.empty:
                continue

            # 构建button上显示的文字
            co_name_str = self._combine_stock_name(stock_info, only_name=True)
            button = widgets.Button(description=co_name_str, disabled=False)
            # 添加一个新属性symbol在button对象里，on_button_clicked使用
            button.symbol = symbol

            def on_button_clicked(bt):
                # 关闭当前整个大的widget界面，重新构建一个全新的界面
                self.widget.close()
                symbol_code = code_to_symbol(bt.symbol)
                self.symbol = symbol_code
                # 重新赋予self.widget值，即一个新的widget
                self.widget = self.stock_base_info(self.symbol)
                # 重新显示新的界面
                self.display()

            button.on_click(on_button_clicked)
            other_co_bs.append(button)

        # 将symbol button一行显示两个，2个为一组，组装子symbol button序列,
        other_co_bs = self._sub_split(other_co_bs, len(other_co_bs) / 2)
        # 将每一组加到一个行box里面
        tbs_boxs = [widgets.HBox(tbs) for tbs in other_co_bs]
        self.other_co_box = widgets.VBox(tbs_boxs)
        return self.other_co_box

    def _sub_split(self, n_buttons, n_split):
        """将symbol button，每n_split个为一组，组装子symbol button序列"""
        sub_bt_cnt = int(len(n_buttons) / n_split)
        if sub_bt_cnt == 0:
            sub_bt_cnt = 1
        group_adjacent = lambda a, k: zip(*([iter(a)] * k))
        bts_group = list(group_adjacent(n_buttons, sub_bt_cnt))
        residue_ind = -(len(n_buttons) % sub_bt_cnt) if sub_bt_cnt > 0 else 0
        if residue_ind < 0:
            bts_group.append(n_buttons[residue_ind:])
        return bts_group

    @catch_error(return_val='')
    def _combine_stock_name(self, stock_info, only_name=False):
        """通过stock_info中的公司信息构建相关字符串名称"""
        if only_name:
            co_name_str = to_unicode(stock_info.co_name.values[0])
        else:
            # eg: 特斯拉电动车（US.NASDQ:TSLA）
            co_name_str = u'{}({}.{}:{})'.format(to_unicode(stock_info.co_name.values[0]),
                                                 to_unicode(stock_info.market.values[0]),
                                                 to_unicode(stock_info.exchange.values[0]),
                                                 to_unicode(stock_info.symbol.values[0]))
        return co_name_str

    def stock_base_info(self, symbol_code):
        """构建股票基本信息：公司简介，业务，市值，市盈率，市净率，每股净资产，流通股本，总股本，机构持股等信息"""
        if not ABuEnv.g_is_ipython:
            logging.info('widget op only support ipython env!')
            return

        stock_info = query_stock_info(symbol_code)
        if stock_info is None or stock_info.empty:
            logging.info('stock_info is None or stock_info.empty!')
            return

        # 公司名称
        co_name_str = self._combine_stock_name(stock_info)
        co_name = widgets.Text(
            value=co_name_str,
            description=u'公司名称:',
            disabled=False
        )

        # 公司简介
        co_intro = None
        if 'co_intro' in stock_info:
            co_intro = widgets.Textarea(
                value=to_unicode(stock_info.co_intro.values[0]),
                description=u'公司简介:',
                disabled=False,
                layout=widgets.Layout(height='226px')
            )

        co_site = None
        if 'co_site' in stock_info:
            site = to_unicode(stock_info.co_site.values[0])
            co_site_str = u'<p><a target="_blank" a href="{}">公司网站: {}</a></p>'.format(
                site, site)
            co_site = widgets.HTML(value=co_site_str)

        pv_dict = {
            'pe_s_d': u"市盈率(静)/(动):",
            'pb_d': u"市净率(动):",
            'pb_MRQ': u"市净率MRQ:",
            'ps_d': u"市销率(动):",
            'ps': u"市销率:",
            'pe_s': u"市盈率(静):"}

        p_widget_tab = self.make_sub_tab_widget(stock_info, pv_dict)

        asset_dict = {
            'mv': u"总市值:",
            'asset': u"每股净资产：",
            'cc': u"流通股本："
        }

        asset_widget_tab = self.make_sub_tab_widget(stock_info, asset_dict)

        equity_dict = {
            'equity': u"总股本:",
            'hk_equity': u"港股股本:",
            'oo': u"机构持股:"
        }
        equity_widget_tab = self.make_sub_tab_widget(stock_info, equity_dict)
        accordion = widgets.Accordion(children=[self.other_industries_symbol(stock_info)])
        industry_str = to_unicode(stock_info.industry.values[0])
        industry = u'行业：{}'.format(industry_str)
        accordion.set_title(0, industry)
        base_info_widgets = list(filter(lambda widget: widget is not None,
                                        [co_name, co_intro, co_site, p_widget_tab, asset_widget_tab, equity_widget_tab,
                                         accordion]))
        base_info = widgets.VBox(base_info_widgets)
        return base_info

    def make_sub_tab_widget(self, stock_info, sub_dict):
        """用于构建：股本/港股股本/机构持股子tab，市盈率/市净率/市销率子tab, 总市值/每股净资产/流通股本子tab"""
        sub_widget_array = []
        sub_widget_table_name = []
        for sc in sub_dict:
            if sc in stock_info.columns:
                sub_name = to_unicode(sub_dict[sc])
                sub_widget = widgets.Text(
                    value=to_unicode(stock_info[sc].values[0]),
                    description=sub_name,
                    disabled=False
                )
                sub_widget_array.append(sub_widget)
                sub_widget_table_name.append(sub_name)

        sub_widget_tab = widgets.Tab()
        sub_widget_tab.children = sub_widget_array
        for ind, name in enumerate(sub_widget_table_name):
            sub_widget_tab.set_title(ind, name)
        return sub_widget_tab



# -*- encoding:utf-8 -*-
"""
    网络统一接口模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging
import requests
import time
import ast

from requests.packages.urllib3.exceptions import ReadTimeoutError

"""g_enable_lru_cache针对lru_cache是否开启，考虑到目前爬取的数据fetch url都会有时间戳等可变字段，所以默认关闭"""
g_enable_lru_cache = False
g_lru_cache_max = 300
if g_enable_lru_cache:
    # 开启import lru_cache
    from ..CoreBu.ABuFixes import lru_cache
else:
    # 导入一个空的装饰器as lru_cache
    from ..UtilBu.ABuDTUtil import empty_wrapper_with_params as lru_cache
# 设置requests库的日志级别
logging.getLogger("requests").setLevel(logging.WARNING)


@lru_cache(maxsize=g_lru_cache_max)
def get(url, params=None, headers=None, retry=3, **kwargs):
    """
    :param url: 请求base url
    :param params: url params参数
    :param headers: http head头信息
    :param retry: 重试次数，默认retry=3
    :param kwargs: 透传给requests.get，可设置ua等，超时等参数
    """
    req_count = 0
    while req_count < retry:
        # 重试retry次
        try:
            resp = requests.get(url=url, params=params, headers=headers, **kwargs)
            if resp.status_code == 200 or resp.status_code == 206:
                # 如果200，206返回，否则继续走重试
                return resp
        except ReadTimeoutError:
            # 超时直接重试就行，不打日志
            pass
        except Exception as e:
            logging.exception(e)
        req_count += 1
        time.sleep(0.5)
        continue
    return None


def post(url, params=None, headers=None, retry=3, **kwargs):
    """
    :param url: 请求base url
    :param params: url params参数
    :param headers: http head头信息
    :param retry: 重试次数，默认retry=3
    :param kwargs: 透传给requests.get，可设置ua等，超时等参数
    """
    req_count = 0
    while req_count < retry:
        try:
            resp = requests.post(url=url, params=params, headers=headers, **kwargs)
            return resp
        except Exception as e:
            logging.exception(e)
            req_count += 1
            time.sleep(0.5)
            continue
    return None


def ast_parse_js(js_var):
    """
    通过ast模块解析Javascript字符串
    :param js_var: Javascript字符串
    :return: map, dict or value
    """
    js_mode = ast.parse(js_var)
    body_head = js_mode.body[0]

    def _parse(node):
        if isinstance(node, ast.Expr):
            return _parse(node.value)
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Dict):
            return dict(zip(map(_parse, node.keys), map(_parse, node.values)))
        elif isinstance(node, ast.List):
            return map(_parse, node.elts)
        else:
            raise NotImplementedError(node.__class__)

    return _parse(body_head)


def parse_js(js_var):
    """
    通过eval解析Javascript字符串
    :param js_var: Javascript字符串
    :return: dict
    """
    obj = eval(js_var, type('Dummy', (dict,), dict(__getitem__=lambda s, n: n))())
    return obj



# -*- encoding:utf-8 -*-
"""
    日内滑点买入示例实现：均价买入
    最简单的回测买入方式，优点简单，且回测高效，在回测交易
    数量足够多的前提下也能接近实盘
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import numpy as np

from .ABuSlippageBuyBase import AbuSlippageBuyBase, slippage_limit_up

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""外部修改默认下跌阀值使用如：abupy.slippage.sbm.g_open_down_rate = 0.02"""
g_open_down_rate = 0.07


class AbuSlippageBuyMean(AbuSlippageBuyBase):
    """示例日内滑点均价买入类"""

    @slippage_limit_up
    def fit_price(self):
        """
        取当天交易日的最高最低均价做为决策价格
        :return: 最终决策的当前交易买入价格
        """

        # TODO 基类提取作为装饰器函数，子类根据需要选择是否装饰，并且添加上根据order的call，put明确细节逻辑
        if self.kl_pd_buy.pre_close == 0 or (self.kl_pd_buy.open / self.kl_pd_buy.pre_close) < (1 - g_open_down_rate):
            # 开盘就下跌一定比例阀值，放弃单子
            return np.inf
        # 买入价格为当天均价，即最高，最低的平均，也可使用高开低收平均等方式计算
        self.buy_price = np.mean([self.kl_pd_buy['high'], self.kl_pd_buy['low']])
        # 返回最终的决策价格
        return self.buy_price



# -*- encoding:utf-8 -*-

"""
MACD

MACD称为指数平滑异动移动平均线，是从双指数移动平均线发展而来的，由快的加权移动均线（EMA12）减去慢的加权移动均线（EMA26）
得到DIF，再用DIF - (快线-慢线的9日加权移动均线DEA）得到MACD柱。MACD的意义和双移动平均线基本相同，即由快、慢均线的离散、
聚合表征当前的多空状态和股价可能的发展变化趋势，但阅读起来更方便。当MACD从负数转向正数，是买的信号。当MACD从正数转向负数，
是卖的信号。当MACD以大角度变化，表示快的移动平均线和慢的移动平均线的差距非常迅速的拉开，代表了一个市场大趋势的转变。

"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .ABuNDBase import plot_from_order, g_calc_type, ECalcType
from ..UtilBu import ABuScalerUtil
from ..UtilBu.ABuDTUtil import catch_error
from ..CoreBu.ABuPdHelper import pd_ewm_mean

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnresolvedReferences
def _calc_macd_from_ta(price, fast_period=12, slow_period=26, signal_period=9):
    """
    使用talib计算macd, 即透传talib.MACD计算结果
    :param price: 收盘价格序列，pd.Series或者np.array
    :param fast_period: 快的加权移动均线线, 默认12，即EMA12
    :param slow_period: 慢的加权移动均线, 默认26，即EMA26
    :param signal_period: dif的指数移动平均线，默认9
    """

    import talib
    if isinstance(price, pd.Series):
        price = price.values

    dif, dea, bar = talib.MACD(price,
                               fastperiod=fast_period,
                               slowperiod=slow_period,
                               signalperiod=signal_period)
    return dif, dea, bar


def _calc_macd_from_pd(price, fast_period=12, slow_period=26, signal_period=9):
    """
    通过macd公式手动计算macd
    :param price: 收盘价格序列，pd.Series或者np.array
    :param fast_period: 快的加权移动均线线, 默认12，即EMA12
    :param slow_period: 慢的加权移动均线, 默认26，即EMA26
    :param signal_period: dif的指数移动平均线，默认9
    """

    if isinstance(price, pd.Series):
        price = price.values

    # 快的加权移动均线
    ewma_fast = pd_ewm_mean(price, span=fast_period)
    # 慢的加权移动均线
    ewma_slow = pd_ewm_mean(price, span=slow_period)
    # dif = 快线 - 慢线
    dif = ewma_fast - ewma_slow
    # dea = dif的9日加权移动均线
    dea = pd_ewm_mean(dif, span=signal_period)
    bar = (dif - dea)
    return dif, dea, bar


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式"""
calc_macd = _calc_macd_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_macd_from_ta


def plot_macd_from_klpd(kl_pd, with_points=None, with_points_ext=None, **kwargs):
    """
    封装plot_macd，绘制收盘价格，macd（dif, dea, bar）曲线
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param kwargs: 绘制技术指标需要的其它关键字参数with_price, fast_period, slow_period, signal_period, 最终透传给plot_macd
    """
    plot_macd(kl_pd.close, kl_pd.index, with_points=with_points, with_points_ext=with_points_ext, **kwargs)


def plot_macd_from_order(order, date_ext=120, **kwargs):
    """
    封装ABuNDBase中的plot_from_order与模块中绘制技术指标的函数，完成技术指标可视化及标注买入卖出点位
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param kwargs: 绘制技术指标需要的其它关键字参数with_price, fast_period, slow_period, signal_period, 最终透传给plot_macd
    """
    return plot_from_order(plot_macd_from_klpd, order, date_ext, **kwargs)


def plot_macd(prices, kl_index, with_points=None, with_points_ext=None,
              with_price=True,
              fast_period=12, slow_period=26, signal_period=9):
    """
    绘制收盘价格，以及对应的macd曲线，如果有with_points点位标注，使用竖线标注
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param kl_index: pd.Index时间序列
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param with_price: 将价格一起绘制，但是价格要做数据标准化
    :param fast_period: 快的加权移动均线线, 默认12，即EMA12
    :param slow_period: 慢的加权移动均线, 默认26，即EMA26
    :param signal_period: dif的指数移动平均线，默认9
    :return:
    """
    dif, dea, bar = calc_macd(
        prices, fast_period=fast_period, slow_period=slow_period, signal_period=signal_period)

    plt.figure(figsize=[14, 7])
    plt.plot(kl_index, dif, label='macd dif')
    plt.plot(kl_index, dea, label='signal dea')

    if with_price:
        plt.plot(kl_index, ABuScalerUtil.scaler_std(prices), label='close price')

    # 绘制红绿macd的bar
    # noinspection PyTypeChecker
    bar_red = np.where(bar > 0, bar, 0)
    # noinspection PyTypeChecker
    bar_green = np.where(bar < 0, bar, 0)
    plt.bar(kl_index, bar_red, facecolor='red', label='hist bar')
    plt.bar(kl_index, bar_green, facecolor='green', label='hist bar')

    @catch_error(return_val=None, log=False)
    def plot_with_point(points, co, cc):
        """
        点位使用圆点＋竖线进行标注
        :param points: 点位坐标序列
        :param co: 点颜色 eg. 'go' 'ro'
        :param cc: markeredgecolor和竖线axvline颜色 eg. 'green' 'red'
        """
        v_index_num = kl_index.tolist().index(points)
        plt.plot(points, dif[v_index_num], co, markersize=12, markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor=cc)
        plt.axvline(points, color=cc)

    # with_points和with_points_ext的点位使用圆点＋竖线标注，plot_with_point方法
    if with_points is not None:
        plot_with_point(with_points, 'go', 'green')

    if with_points_ext is not None:
        plot_with_point(with_points_ext, 'ro', 'red')

    plt.grid(True)
    plt.legend(loc='best')
    plt.show()



# -*- encoding:utf-8 -*-
"""
    风险控制仓位管理基础
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABCMeta, abstractmethod

from ..CoreBu.ABuFixes import six
from ..MarketBu.ABuMarket import MarketMixin

"""每一笔交易最大仓位比例设置，外部可通过如：abupy.beta.position.g_pos_max = 0.5修改最大每一笔交易最大仓位比例"""
g_pos_max = 0.75
"""
    保证净最小比例，默认1，即不使用融资，不会触发Margin Call
    在期货数据中有每种商品最少保证金比例，可使用设置
    外部可通过如：abupy.beta.position.g_deposit_rate = 0.5
"""
g_deposit_rate = 1


class AbuPositionBase(six.with_metaclass(ABCMeta, MarketMixin)):
    """仓位管理抽象基类"""

    def __init__(self, kl_pd_buy, factor_name, symbol_name, bp, read_cash, deposit_rate):
        """
        :param kl_pd_buy: 交易当日的交易数据
        :param factor_name: 因子名称
        :param symbol_name: symbol代码
        :param bp: 买入价格
        :param read_cash: 初始资金
        :param deposit_rate: 保证金比例
        """
        self.kl_pd_buy = kl_pd_buy
        self.factor_name = factor_name
        self.symbol_name = symbol_name
        self.bp = bp
        self.read_cash = read_cash
        self.deposit_rate = deposit_rate

    def __str__(self):
        """打印对象显示：class name, factor_name, symbol_name, read_cash, deposit_rate"""
        return '{}: factor_name:{}, symbol_name:{}, read_cash:{}, deposit_rate:{}'.format(self.__class__.__name__,
                                                                                          self.factor_name,
                                                                                          self.symbol_name,
                                                                                          self.read_cash,
                                                                                          self.deposit_rate)

    __repr__ = __str__

    @abstractmethod
    def fit_position(self, factor_object):
        """
        fit_position计算的结果是买入多少个单位（股，手，顿，合约）具体计算子类实现
        :param factor_object: ABuFactorBuyBases实例对象
        :return:买入多少个单位（股，手，顿，合约）
        """
        pass



# -*- encoding:utf-8 -*-
"""mac os 平台工具模块"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import socket


def socket_bind_recv(socket_fn, cmd_handler):
    """
    基于bsd系统的进程间socket通信，接受消息，处理消息
    :param socket_fn: socket文件名称
    :param cmd_handler: cmd处理函数，callable类型
    """
    if not callable(cmd_handler):
        print('socket_bind_recv cmd_handler must callable!')

    server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    server.bind(socket_fn)
    server.listen(0)
    while True:
        connection, _ = server.accept()
        socket_cmd = connection.recv(1024).decode()
        # 把接收到的socket传递给外部对应的处理函数
        cmd_handler(socket_cmd)
        connection.close()


def socket_send_msg(socket_fn, msg):
    """
    基于bsd系统的进程间socket通信，发送消息
    :param socket_fn: : socket文件名称
    :param msg: 字符串类型需要传递的数据，不需要encode，内部进行encode
    """
    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client.connect(socket_fn)
    client.send(msg.encode())
    client.close()


def show_msg(title, msg):
    """
    使用osascript脚步提示弹窗，主要用在长时间且耗时的任务中，提示重要问题信息
    :param title: 弹窗标题
    :param msg: 弹窗信息
    """
    msg_cmd = 'osascript -e \'display notification "%s" with title "%s"\'' % (msg, title)
    os.system(msg_cmd)


def fold_free_size_mb(folder):
    """
    mac os下剩余磁盘空间获取
    :param folder: 目标目录
    :return: 返回float，单位mb
    """
    st = os.statvfs(folder)
    return st.f_bavail * st.f_frsize / 1024 / 1024



# -*- encoding:utf-8 -*-
"""
    技术线对象，对外执行，输出模块
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools
import logging
import math
from collections import Iterable
from enum import Enum

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d

from ..TLineBu.ABuTLExecute import shift_distance, bfgs_min_pos, support_resistance_pos, \
    select_k_support_resistance, plot_support_resistance_trend, support_resistance_predict, regress_trend_channel, \
    below_above_gen, find_percent_point, find_golden_point_ex, find_golden_point, skeleton_how

from ..CoreBu import ABuEnv
from ..CoreBu.ABuBase import FreezeAttrMixin
from ..UtilBu import ABuRegUtil
from ..UtilBu import ABuStatsUtil
from ..UtilBu.ABuDTUtil import arr_to_numpy
from ..UtilBu.ABuLazyUtil import LazyFunc

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""模块打印根据环境选择logging.info或者print函数"""
log_func = logging.info if ABuEnv.g_is_ipython else print

"""预备颜色序列集，超出序列数量应使用itertools.cycle循环绘制"""
K_PLT_MAP_STYLE = [
    'r', 'g', 'c', 'k', 'm', 'r', 'y']


class ESkeletonHow(Enum):
    """计算骨架走势使用的how"""

    """使用最小值取骨架点位"""
    skeleton_min = 0
    """使用最大值取骨架点位"""
    skeleton_max = 1
    """使用平均值取骨架点位"""
    skeleton_mean = 2
    """使用中位数取骨架点位"""
    skeleton_median = 3

    """使用时间序列最后的元素取骨架点位"""
    skeleton_close = 4
    """
        使用三角模式，即最高，最低，第三点：
        确定取最大值，最小值，第三个点位how_func提供，
        如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，
        第三点取序列起点，否则取序列终点
    """
    skeleton_triangle = 100


class EShiftDistanceHow(Enum):
    """计算位移路程比的how"""

    """
        使用时间序列最后的元素做为路程的计算基础：
        对应序列的最后一个点位值，标准路程点位值定义
    """
    shift_distance_close = 0
    """
        使用极限值做为路程的计算基础：
        如果p_arr[0] > p_arr[-1]，使用np.min(p_arr)，否则np.max(p_arr)，即上升趋势取max，下跌趋势取min
    """
    shift_distance_maxmin = 1
    """
        使用序列的sum和极限值为路程的计算基础：
        如果abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) 取np.min(p_arr)否则np.max(p_arr)
    """
    shift_distance_sum_maxmin = 2


"""step_x_to_step函数中序列步长的常数单元值"""
g_step_unit = 10


class AbuTLine(FreezeAttrMixin):
    """技术线封装执行对外操作的对象类"""

    def __init__(self, line, line_name, **kwargs):
        """
        :param line: 技术线可迭代序列，内部会通过arr_to_numpy统一转换numpy
        :param line_name: 技术线名称，str对象
        :param kwargs mean: 外部可选择通过kwargs设置mean，如不设置line.mean()
        :param kwargs std: 外部可选择通过kwargs设置std，如不设置line.std()
        :param kwargs high: 外部可选择通过kwargs设置high，如不设置self.mean + self.std
        :param kwargs low: 外部可选择通过kwargs设置low，如不设置self.mean - self.std
        :param kwargs close: 外部可选择通过kwargs设置close，如不设置line[-1]
        """

        # 把序列的nan进行填充，实际上应该是外面根据数据逻辑把nan进行填充好了再传递进来，这里只能都使用bfill填了
        line = pd.Series(line).fillna(method='bfill')
        self.tl = arr_to_numpy(line)
        self.mean = kwargs.pop('mean', self.tl.mean())
        self.std = kwargs.pop('std', self.tl.std())
        self.high = kwargs.pop('high', self.mean + self.std)
        self.low = kwargs.pop('low', self.mean - self.std)
        self.close = kwargs.pop('close', self.tl[-1])

        self.x = np.arange(0, self.tl.shape[0])
        self.line_name = line_name

        for k, v in kwargs:
            # 需要设置什么都通过kwargs设置进来，不然_freeze后无法设置
            setattr(self, k, v)
        # 需要进行定稿，初始化好就不能动
        self._freeze()

    @classmethod
    def show_kl_pd(cls, kl_pd, key='close', show=True, **kwargs):
        """
        类方法，针对金融时间序列中的数据列进行技术线分析，可视化最优拟合次数，
        路程位移比，拟合通道曲线，骨架通道，阻力位和支撑位等技术线分析，返回
        AbuTLine对象
        :param kl_pd: 金融时间序列，pd.DataFrame对象
        :param key: kl_pd中做为技术线的列名称，str对象
        :param show: 是否可视化，可视化最优拟合次数，路程位移比，拟合通道曲线，骨架通道，阻力位和支撑位等
        :param kwargs: 可视化函数涉及的其它参数 eg：step_x, only_last等
        :return: 返回AbuTLine对象
        """
        line = cls(kl_pd[key], key)
        if show:
            # 可以通过kwargs设置show的参数，先pop出来
            zoom = kwargs.pop('zoom', False)
            step_x = kwargs.pop('step_x', 1.0)
            how = kwargs.pop('how', EShiftDistanceHow.shift_distance_close)
            only_last = kwargs.pop('only_last', False)

            line.show()
            # 可视化技术线最优拟合次数
            line.show_best_poly(zoom=zoom)
            # 可视化技术线'路程位移比'
            line.show_shift_distance(step_x=step_x, how=how)
            # 可视化技术线拟合曲线及上下拟合通道曲线
            line.show_regress_trend_channel(step_x=step_x)
            # 可视化可视化技术线骨架通道
            line.show_skeleton_channel(step_x=step_x)
            # 可视化技术线比例分割的区域
            line.show_percents()
            # 可视化技术线黄金分割
            line.show_golden()
            # 对技术线阻力位和支撑位进行绘制, 以及所有中间过程
            line.show_support_resistance_trend(only_last=only_last)
        return line

    @LazyFunc
    def score(self):
        """
        被LazyFunc装饰：
        score代表当前技术线值在当前的位置， (self.close - self.low) / (self.high - self.low)
        eg：
            self.high ＝ 100， self.low＝0，self.close＝80
            －> (self.close - self.low) / (self.high - self.low) = 0.8
            即代表当前位置在整体的0.8位置上

        :return: 技术线当前score, 返回值在0-1之间
        """
        if self.high == self.low:
            score = 0.8 if self.close > self.low else 0.2
        else:
            score = (self.close - self.low) / (self.high - self.low)
        return score

    @LazyFunc
    def y_zoom(self):
        """
        被LazyFunc装饰：
        获取对象技术线tl被self.x缩放后的序列y_zoom
        :return: 放后的序列y_zoom
        """
        zoom_factor = self.x.max() / self.tl.max()
        y_zoom = zoom_factor * self.tl
        return y_zoom

    def step_x_to_step(self, step_x):
        """
        针对技术线的时间范围步长选择函数，在show_shift_distance，show_regress_trend_channel，
        show_skeleton_channel等涉及时间步长的函数中用来控制步长范围
        :param step_x: 时间步长控制参数，float
        :return: 最终输出被控制在2-len(self.tl), int
        """

        if step_x <= 0:
            # 不正常step_x规范到正常范围中
            log_func('input step_x={} is error, change to step_x=1'.format(step_x))
            step_x = 1
        # 如果需要调整更细的粒度，调整g_step_unit的值
        step = int(math.floor(len(self.tl) / g_step_unit / step_x))
        # 输出被控制在2-len(self.tl)
        step = len(self.tl) if step > len(self.tl) else step
        step = 2 if step < 2 else step
        return step

    def show(self):
        """可视化技术线最基本的信息，high，mean，low"""
        plt.subplots()
        # tl装载技术线本体
        plt.plot(self.tl)
        plt.axhline(self.high, color='c')
        plt.axhline(self.mean, color='r')
        plt.axhline(self.low, color='g')
        _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)
        plt.legend(['TLine', 'high', 'mean', 'low'],
                   bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.title(self.line_name)
        plt.show()

    def show_best_poly(self, zoom=False, show=show):
        """
        可视化技术线最优拟合次数，寻找poly（1－100）次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势，
        具体详情ABuRegUtil.search_best_poly
        :param zoom: 透传search_best_poly是否缩放x,y
        :param show: 是否进行可视化
        """
        best = ABuRegUtil.search_best_poly(self.tl, zoom=zoom, show=show)
        if show:
            log_func('best poly = {}, zoom={}'.format(best, zoom))
        return best

    def show_least_valid_poly(self, zoom=False, show=True):
        """
        可视化技术线，检测至少poly次拟合曲线可以代表原始曲线y的走势，
        具体详情ABuRegUtil.least_valid_poly
        :param zoom: 透传least_valid_poly是否缩放x,y
        :param show: 是否进行可视化
        """
        least = ABuRegUtil.least_valid_poly(self.tl, zoom=zoom, show=show)
        if show:
            log_func('least poly = {}, zoom={}'.format(least, zoom))
        return least

    def show_shift_distance(self, how=EShiftDistanceHow.shift_distance_close, step_x=1.0, show=True, show_log=True):
        """
        可视化技术线'路程位移比'，注意默认使用shift_distance_close对应标准路程点位值定义方法，其它方法对应的
        路程终点点位值使用的计算方法并非得到最准确的'路程位移比'，实现详ABuTLExecute.shift_distance
        :param how: EShiftDistanceHow对象或者callable即外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
        :param step_x: 时间步长控制参数，默认1.0，float
        :param show: 是否进行可视化
        :param show_log: 是否输出位移路程比各个段比值，默认True
        :return 对每一个金融序列切片进行shift_distance的返回结果序列，即每一个序列中的元素为：
                    h_distance(三角底边距离), v_distance(三角垂直距离),
                    distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
                所组成的tuple对象
        """

        # 这里使用了缩放后的y，因为确保路程位移比具有更好的适应性
        y = self.y_zoom
        step = self.step_x_to_step(step_x)

        if show:
            plt.plot(y)
        shift_distance_list = []
        for slice_end, color in zip(np.arange(step, len(y), step), itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            shift_distance_list.append(
                shift_distance(y, how, slice_start=slice_start, slice_end=slice_end,
                               color=color, show=show, show_log=show_log, ps=False))
        if show:
            plt.show()
        return shift_distance_list

    def show_regress_trend_channel(self, step_x=1.0):
        """
        可视化技术线拟合曲线及上下拟合通道曲线，返回三条拟合曲线，组成拟合通道
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        y = self.tl
        step = self.step_x_to_step(step_x)

        plt.plot(y)
        for slice_end, color in zip(np.arange(step, len(y), step), itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            slice_arr = y[slice_start:slice_end]
            # 通过regress_trend_channel获取切片段的上中下三段拟合曲线值
            y_below, y_fit, y_above = regress_trend_channel(slice_arr)
            x = self.x[slice_start:slice_end]
            plt.plot(x, y_below, 'g')
            plt.plot(x, y_fit, 'y')
            plt.plot(x, y_above, 'r')
        plt.show()

    def show_skeleton_channel(self, with_mean=True, step_x=1.0):
        """
        套接show_skeleton，可视化可视化技术线骨架通道，通道由：
        ESkeletonHow.skeleton_min：下通道，
        ESkeletonHow.skeleton_max：上通道，
        ESkeletonHow.skeleton_mean 中轨通道，组成

        :param with_mean: 是否绘制ESkeletonHow.skeleton_mean 中轨通道，默认True
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        self.show_skeleton(how=ESkeletonHow.skeleton_min, step_x=step_x, ps=False)
        self.show_skeleton(how=ESkeletonHow.skeleton_max, step_x=step_x, ps=False)
        if with_mean:
            self.show_skeleton(how=ESkeletonHow.skeleton_mean, step_x=step_x, ps=False)
        plt.plot(self.tl)
        # 前面的绘制ps都是False, 这里统一show
        plt.show()

    def show_skeleton(self, how=ESkeletonHow.skeleton_min, step_x=1.0, ps=True):
        """
        可视化技术线骨架结构
        :param how: 计算数据序列骨架点位的方法，ESkeletonHow对象或者callable即外部可自行设置方法，
                    即计算数据序列骨架点位的方法可自定义
        :param step_x: 时间步长控制参数，默认1.0，float
        :param ps: 是否立即执行plt.show()
        """
        step = self.step_x_to_step(step_x)

        # 每个单位都先画一个点，由两个点连成一条直线形成股价骨架图
        last_pos = None
        # 根据how映射计算数据序列骨架点位的方法
        how_func = skeleton_how(how)
        for slice_end, color in zip(np.arange(step, len(self.tl), step), itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            slice_arr = self.tl[slice_start:slice_end]
            if how == ESkeletonHow.skeleton_triangle:
                """
                    三角模式骨架点位：确定取最大值，最小值，第三个点位how_func提供
                    如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，第三点取序列起点，否则取序列终点
                """
                max_pos = (np.argmax(slice_arr) + slice_start, np.max(slice_arr))
                min_pos = (np.argmin(slice_arr) + slice_start, np.min(slice_arr))
                draw_pos = how_func(slice_arr, slice_start)

                plt.plot([draw_pos[0], min_pos[0]],
                         [draw_pos[1], min_pos[1]], c=color)

                plt.plot([draw_pos[0], max_pos[0]],
                         [draw_pos[1], max_pos[1]], c=color)

                plt.plot([min_pos[0], max_pos[0]],
                         [min_pos[1], max_pos[1]], c=color)
            else:
                # 其它骨架数据计算方法
                draw_pos = (slice_start, how_func(slice_arr))
                if last_pos is not None:
                    # 将两两临近局部最小值相连，两个点连成一条直线
                    plt.plot([last_pos[0], draw_pos[0]],
                             [last_pos[1], draw_pos[1]], 'o-')
                # 将这个步长单位内的最小值点赋予last_pos
                last_pos = draw_pos
        if ps:
            plt.plot(self.tl)
            plt.show()

    def show_skeleton_bfgs(self, step_x=1.0):
        """
        可视化技术线骨架结构与show_skeleton不同，由bfgs确定骨架点位值，即通过
        scipy.interpolate.interp1d插值形成模型通过sco.fmin_bfgs计算骨架点位值
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        step = self.step_x_to_step(step_x)
        # scipy.interpolate.interp1d插值形成模型
        linear_interp = interp1d(self.x, self.tl)
        # 每个单位都先画一个点，由两个点连成一条直线形成股价骨架图
        last_pos = None

        # 每步长step单位求一次局部最小
        for find_min_pos in np.arange(step, len(self.tl), step):
            # sco.fmin_bfgs计算骨架点位值
            local_min_pos = int(bfgs_min_pos(find_min_pos, linear_interp, len(self.tl)))
            if local_min_pos == -1:
                # 其实主要就是利用这里找不到的情况进行过滤
                continue

            # 形成最小点位置信息(x, y)
            draw_pos = (local_min_pos, self.tl[local_min_pos])
            # 第一个step单位last_pos＝none, 之后都有值
            if last_pos is not None:
                # 将两两临近局部最小值相连，两个点连成一条直线
                plt.plot([last_pos[0], draw_pos[0]],
                         [last_pos[1], draw_pos[1]], 'o-')
            # 将这个步长单位内的最小值点赋予last_pos
            last_pos = draw_pos
        plt.show()

    def show_support_resistance_pos(self, best_poly=0, show=True):
        """
        可视化分析技术线阻力位和支撑位，通过sco.fmin_bfgs寻找阻力位支撑位，阻力位点也是通过sco.fmin_bfgs寻找，
        但是要求传递进来的序列已经是标准化后取反的序列
        eg：
            demean_y = ABuStatsUtil.demean(self.tl)： 首先通过demean将序列去均值
            resistance_y = demean_y * -1 ：阻力位序列要取反
            support_y = demean_y ：支持位序列不需要取反
        :param best_poly: 函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了,
                          详细阅ABuTLExecute.support_resistance_pos
        :param show: 是否可视化
        :return: (技术线支撑位: support_pos, 技术线阻力位: resistance_pos)
        """

        # 首先通过demean将序列去均值
        demean_y = ABuStatsUtil.demean(self.tl)
        # 阻力位序列要取反
        resistance_y = demean_y * -1
        # 支持位序列不需要取反
        support_y = demean_y

        # 分析技术线支撑位
        support_pos = support_resistance_pos(self.x, support_y, best_poly=best_poly, label='support pos')
        # 分析技术线阻力位
        resistance_pos = support_resistance_pos(self.x, resistance_y, best_poly=best_poly,
                                                label='resistance pos')
        if show:
            plt.plot(self.x, self.tl, '--', support_pos, self.tl[support_pos], 'o',
                     resistance_pos, self.tl[resistance_pos], 'p')
            plt.show()
        # 返回 (技术线支撑位: support_pos, 技术线阻力位: resistance_pos)
        return support_pos, resistance_pos

    def show_support_resistance_select_k(self, best_poly=0, show=True):
        """
        可视化分析技术线阻力位和支撑位序列从1-序列个数开始聚类，多个聚类器的方差值进行比较，
        通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param show: 是否可视化显示
        :return: upport_est, resistance_est, support_pos, resistance_pos
        """
        # 可视化分析技术线阻力位或者支撑位
        support_pos, resistance_pos = self.show_support_resistance_pos(best_poly, show=show)
        support_pos = np.array([support_pos, [self.tl[support] for support in support_pos]]).T
        resistance_pos = np.array([resistance_pos, [self.tl[resistance] for resistance in resistance_pos]]).T

        support_est = None
        if len(support_pos) > 1:
            # 多个聚类器的方差值进行比较，通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
            # 注意这里的show直接False了
            support_est = select_k_support_resistance(support_pos, label='support k choice', show=False)

        resistance_est = None
        if len(resistance_pos) > 1:
            # 注意这里的show直接False了
            resistance_est = select_k_support_resistance(resistance_pos, label='resistance k choice', show=False)

        return support_est, resistance_est, support_pos, resistance_pos

    def show_support_resistance_trend(self, best_poly=0, only_last=False, plot_org=False, show=True, show_step=False):
        """
        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        最终对技术线阻力位和支撑位进行绘制，注意show参数控制的是中间流程中的可视化，不包括
        最终阻力位和支撑的可视化
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        :param plot_org: 透传ABuTLExecute.plot_support_resistance_trend，控制是否绘制线段还是直线，控制是否绘制线段还是直线，
                         plot_org＝True时绘制线段，否则通过LinearRegression进行
        :param show_step: show_step参数控制的是中间流程中的可视化, 不包括最终阻力位或者支撑的可视化
        :param show: show: show参数控制的是最终阻力位或者支撑的可视化
        """
        support_est, resistance_est, support_pos, resistance_pos = self.show_support_resistance_select_k(best_poly,
                                                                                                         show=show_step)

        y_trend_dict = {}
        if support_est is not None:
            # FIXME 针对极端没有找到足够绘制支撑阻力位的情况做处理
            support_trend = support_resistance_predict(self.x, self.tl, support_est, support_pos,
                                                       is_support=True, show=show_step)
            y_support_trend = plot_support_resistance_trend(self.x, self.tl, support_trend, 'support trend line',
                                                            only_last=only_last, plot_org=plot_org, show=show)
            if y_support_trend is not None:
                y_trend_dict['support'] = y_support_trend
        else:
            log_func('can\'t plot support !')

        if resistance_est is not None:
            resistance_trend = support_resistance_predict(self.x, self.tl, resistance_est,
                                                          resistance_pos, is_support=False, show=show_step)
            y_resistance_trend = plot_support_resistance_trend(self.x, self.tl, resistance_trend,
                                                               'resistance trend line',
                                                               only_last=only_last, plot_org=plot_org, show=show)
            if y_resistance_trend is not None:
                y_trend_dict['resistance'] = y_resistance_trend
        else:
            log_func('can\'t plot resistance !')

        if show:
            plt.legend(loc=2)
            plt.show()
        return y_trend_dict

    def show_support_trend(self, best_poly=0, only_last=False, plot_org=False, show=True, show_step=False):
        """
        最终对技术线只对阻力位进行绘制

        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        :param plot_org: 透传ABuTLExecute.plot_support_resistance_trend，控制是否绘制线段还是直线，控制是否绘制线段还是直线，
                         plot_org＝True时绘制线段，否则通过LinearRegression进行
        :param show_step: show_step参数控制的是中间流程中的可视化, 不包括最终阻力位或者支撑的可视化
        :param show: show: show参数控制的是最终阻力位或者支撑的可视化
        """

        support_est, _, support_pos, _ = self.show_support_resistance_select_k(best_poly,
                                                                               show=show_step)
        y_trend_dict = {}
        if support_est is not None:
            support_trend = support_resistance_predict(self.x, self.tl, support_est, support_pos,
                                                       is_support=True, show=show_step)
            y_support_trend = plot_support_resistance_trend(self.x, self.tl, support_trend, 'support trend line',
                                                            only_last=only_last, plot_org=plot_org, show=show)
            if y_support_trend is not None:
                y_trend_dict['support'] = y_support_trend
        if show:
            plt.legend(loc=2)
            plt.show()
        return y_trend_dict

    def show_resistance_trend(self, best_poly=0, only_last=False, plot_org=False, show=True, show_step=False):
        """
        最终对技术线只对支撑位进行绘制

        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        :param plot_org: 透传ABuTLExecute.plot_support_resistance_trend，控制是否绘制线段还是直线，控制是否绘制线段还是直线，
                         plot_org＝True时绘制线段，否则通过LinearRegression进行
        :param show_step: show_step参数控制的是中间流程中的可视化, 不包括最终阻力位或者支撑的可视化
        :param show: show: show参数控制的是最终阻力位或者支撑的可视化
        """
        _, resistance_est, _, resistance_pos = self.show_support_resistance_select_k(best_poly,
                                                                                     show=show_step)

        y_trend_dict = {}
        if resistance_est is not None:
            resistance_trend = support_resistance_predict(self.x, self.tl, resistance_est,
                                                          resistance_pos, is_support=False, show=show_step)
            y_resistance_trend = plot_support_resistance_trend(self.x, self.tl, resistance_trend,
                                                               'resistance trend line',
                                                               only_last=only_last, plot_org=plot_org, show=show)
            if y_resistance_trend is not None:
                y_trend_dict['resistance'] = y_resistance_trend
        if show:
            plt.legend(loc=2)
            plt.show()
        return y_trend_dict

    def show_percents(self, percents=(0.1, 0.9)):
        """
        可视化技术线比例分割的区域
        :param percents: float值或者可迭代序列，默认使用(0.1, 0.9)
        :return:
        """
        if not isinstance(percents, Iterable):
            # 如果不是可迭代序列，添加到list中，便于统一处理
            percents = [percents]

        pts_dict = find_percent_point(percents, self.tl)
        plt.plot(self.tl)

        """
            eg: pts_dict 形如：
            {0.1: (15.732749999999999, 15.5075), 0.9: (31.995000000000005, 34.387500000000003)}
            即返回的是一个比例地带，绘制地带的上下边界
        """
        for pt, color in zip(pts_dict, itertools.cycle(K_PLT_MAP_STYLE)):
            below, above = below_above_gen(*pts_dict[pt])
            plt.axhline(below, c=color)
            plt.axhline(above, c=color)

            plt.fill_between(self.x, below, above,
                             alpha=0.5, color=color)
        plt.show()

    def show_golden(self, both_golden=True):
        """
        可视化技术线黄金分割
        :param both_golden: 代表同时可视化两种分割线的计算在一个画布上
        :return:
        """
        if both_golden:
            # 同时可视化两种分割线的计算在一个画布上直接套接show_percents
            self.show_percents(percents=(0.382, 0.618))
        else:
            # 分别可视化
            find_golden_point_ex(self.x, self.tl, show=True)
            find_golden_point(self.x, self.tl, show=True)

    def __str__(self):
        """打印对象显示：line_name: close， below， above, mean"""
        return "{}: now:{} below:{} above:{}".format(self.line_name,
                                                     self.close,
                                                     self.low,
                                                     self.high, self.mean)

    __repr__ = __str__



# -*- encoding:utf-8 -*-
"""选股因子参数以及选择图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ipywidgets as widgets
import numpy as np

from ..PickStockBu.ABuPickRegressAngMinMax import AbuPickRegressAngMinMax
from ..PickStockBu.ABuPickStockDemo import AbuPickStockShiftDistance, AbuPickStockNTop
from ..PickStockBu.ABuPickStockPriceMinMax import AbuPickStockPriceMinMax
from ..WidgetBu.ABuWGPSBase import WidgetPickStockBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class PSPriceWidget(WidgetPickStockBase):
    """对应AbuPickStockPriceMinMax策略widget"""

    def _init_widget(self):
        """构建AbuPickStockPriceMinMax策略参数界面"""

        self.description = widgets.Textarea(
            value=u'价格选股因子策略：\n'
                  u'根据交易目标的一段时间内收盘价格的最大，最小值进行选股，选中规则：\n'
                  u'1. 交易目标最小价格 > 最小价格阀值\n'
                  u'2. 交易目标最大价格 < 最大价格阀值\n',
            description=u'价格选股',
            disabled=False,
            layout=self.description_layout
        )

        self.price_min_label = widgets.Label(u'设定选股价格最小阀值，默认15', layout=self.label_layout)
        self.price_min_float = widgets.FloatText(
            value=15,
            description=u'最小:',
            disabled=False
        )
        self.price_min_ck = widgets.Checkbox(
            value=True,
            description=u'使用最小阀值',
            disabled=False
        )

        def price_min_ck_change(change):
            self.price_min_float.disabled = not change['new']

        self.price_min_ck.observe(price_min_ck_change, names='value')
        self.price_min_box = widgets.VBox([self.price_min_label, self.price_min_ck, self.price_min_float])

        self.price_max_label = widgets.Label(u'设定选股价格最大阀值，默认50', layout=self.label_layout)
        self.price_max_float = widgets.FloatText(
            value=50,
            description=u'最大:',
            disabled=False
        )
        self.price_max_ck = widgets.Checkbox(
            value=True,
            description=u'使用最大阀值',
            disabled=False
        )

        def price_max_ck_change(change):
            self.price_max_float.disabled = not change['new']

        self.price_max_ck.observe(price_max_ck_change, names='value')
        self.price_max_box = widgets.VBox([self.price_max_label, self.price_max_ck, self.price_max_float])
        self.widget = widgets.VBox([self.description, self.price_min_box, self.price_max_box,
                                    self.xd_box, self.reversed_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def make_pick_stock_unique(self):
        """对应按钮添加AbuPickStockPriceMinMax策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        price_min = self.price_min_float.value if self.price_min_ck.value else -np.inf
        price_max = self.price_max_float.value if self.price_max_ck.value else np.inf

        factor_dict = {'class': AbuPickStockPriceMinMax,
                       'xd': self.xd.value,
                       'reversed': self.reversed.value,
                       'threshold_price_min': price_min,
                       'threshold_price_max': price_max}

        factor_desc_key = u'价格选股最大:{}最小:{},周期:{},反转:{}'.format(
            price_max, price_min, self.xd.value, self.reversed.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuPickStockPriceMinMax"""
        return AbuPickStockPriceMinMax


class PSRegressAngWidget(WidgetPickStockBase):
    """对应AbuPickRegressAngMinMax策略widget"""

    def _init_widget(self):
        """构建AbuPickRegressAngMinMax策略参数界面"""

        self.description = widgets.Textarea(
            value=u'拟合角度选股因子策略：\n'
                  u'将交易目标前期走势进行线性拟合计算一个角度，选中规则：\n'
                  u'1. 交易目标前期走势拟合角度 > 最小拟合角度\n'
                  u'2. 交易目标前期走势拟合角度 < 最大拟合角度\n',
            description=u'角度选股',
            disabled=False,
            layout=self.description_layout
        )

        self.ang_min_label = widgets.Label(u'设定选股角度最小阀值，默认-5', layout=self.label_layout)
        self.ang_min_float = widgets.IntText(
            value=-5,
            description=u'最小:',
            disabled=False
        )
        self.ang_min_ck = widgets.Checkbox(
            value=True,
            description=u'使用最小阀值',
            disabled=False
        )

        def ang_min_ck_change(change):
            self.ang_min_float.disabled = not change['new']

        self.ang_min_ck.observe(ang_min_ck_change, names='value')
        self.ang_min_box = widgets.VBox([self.ang_min_label, self.ang_min_ck, self.ang_min_float])

        self.ang_max_label = widgets.Label(u'设定选股角度最大阀值，默认5', layout=self.label_layout)
        self.ang_max_float = widgets.IntText(
            value=5,
            description=u'最大:',
            disabled=False
        )
        self.ang_max_ck = widgets.Checkbox(
            value=True,
            description=u'使用最大阀值',
            disabled=False
        )

        def ang_max_ck_change(change):
            self.ang_max_float.disabled = not change['new']

        self.ang_max_ck.observe(ang_max_ck_change, names='value')
        self.ang_max_box = widgets.VBox([self.ang_max_label, self.ang_max_ck, self.ang_max_float])

        self.widget = widgets.VBox([self.description, self.ang_min_box, self.ang_max_box,
                                    self.xd_box, self.reversed_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def make_pick_stock_unique(self):
        """对应按钮添加AbuPickRegressAngMinMax策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        ang_min = self.ang_min_float.value if self.ang_min_ck.value else -np.inf
        ang_max = self.ang_max_float.value if self.ang_max_ck.value else np.inf
        factor_dict = {'class': AbuPickRegressAngMinMax,
                       'xd': self.xd.value,
                       'reversed': self.reversed.value,
                       'threshold_ang_min': ang_min,
                       'threshold_ang_max': ang_max}

        factor_desc_key = u'角度选股最大:{}最小:{},周期:{},反转:{}'.format(
            ang_max, ang_min, self.xd.value, self.reversed.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuPickRegressAngMinMax"""
        return AbuPickRegressAngMinMax


class PSShiftDistanceWidget(WidgetPickStockBase):
    """对应AbuPickStockShiftDistance策略widget"""

    def _init_widget(self):
        """构建AbuPickStockShiftDistance策略参数界面"""

        self.description = widgets.Textarea(
            value=u'位移路程比选股因子策略：\n'
                  u'将交易目标走势每月计算价格位移路程比，根据比值进行选股，选取波动程度不能太大，也不太小的目标：\n'
                  u'1. 定义位移路程比大于参数阀值的月份为大波动月\n'
                  u'2. 一年中大波动月数量 < 最大波动月个数\n'
                  u'3. 一年中大波动月数量 > 最小波动月个数\n',
            description=u'位移路程',
            disabled=False,
            layout=self.description_layout
        )

        threshold_sd_label1 = widgets.Label(u'设大波动位移路程比阀值，期货市场建议2.0及以上', layout=self.label_layout)
        threshold_sd_label2 = widgets.Label(u'设大波动位移路程比阀值，股票市场建议3.0及以上', layout=self.label_layout)
        self.threshold_sd = widgets.FloatSlider(
            value=2.0,
            min=1.0,
            max=6.0,
            step=0.1,
            description=u'阀值',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
        )
        self.threshold_sd_box = widgets.VBox([threshold_sd_label1, threshold_sd_label2, self.threshold_sd])

        max_cnt_label = widgets.Label(u'选取大波动月数量 < 下面设定的最大波动月个数', layout=self.label_layout)
        min_cnt_label = widgets.Label(u'选取大波动月数量 > 下面设定的最小波动月个数', layout=self.label_layout)
        self.min_max_range = widgets.IntRangeSlider(
            value=[1, 4],
            min=0,
            max=10,
            step=1,
            description=u'范围',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d',
        )
        self.min_max_box = widgets.VBox([max_cnt_label, min_cnt_label, self.min_max_range])
        # 这个策略不要可自定义xd，限定选股周期为1年
        self.widget = widgets.VBox([self.description, self.threshold_sd_box,
                                    self.min_max_range, self.reversed_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def make_pick_stock_unique(self):
        """对应按钮添加AbuPickStockShiftDistance策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""

        factor_dict = {'class': AbuPickStockShiftDistance,
                       'threshold_sd': self.threshold_sd.value,
                       'reversed': self.reversed.value,
                       'threshold_max_cnt': self.min_max_range.value[1],
                       'threshold_min_cnt': self.min_max_range.value[0]}

        factor_desc_key = u'位移路程选股大波动:{}最大:{}最小:{},反转:{}'.format(
            self.threshold_sd.value, self.min_max_range.value[1], self.min_max_range.value[0], self.reversed.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuPickStockShiftDistance"""
        return AbuPickStockShiftDistance


class PSNTopWidget(WidgetPickStockBase):
    """对应AbuPickStockNTop策略widget"""

    def _init_widget(self):
        """构建AbuPickStockNTop策略参数界面"""
        self.description = widgets.Textarea(
            value=u'涨跌幅top N选股因子策略：\n'
                  u'选股周期上对多只股票涨跌幅进行排序，选取top n个股票做为交易目标：\n'
                  u'(只对在股池中选定的symbol序列生效，对全市场回测暂时不生效)\n',
            description=u'top N涨跌',
            disabled=False,
            layout=self.description_layout
        )

        n_top_label = widgets.Label(u'设定选取top个交易目标数量，默认3', layout=self.label_layout)
        self.n_top = widgets.IntText(
            value=3,
            description=u'TOP N',
            disabled=False
        )
        self.n_top_box = widgets.VBox([n_top_label, self.n_top])

        direction_top_label1 = widgets.Label(u'direction_top参数的意义为选取方向：', layout=self.label_layout)
        direction_top_label2 = widgets.Label(u'默认值为正：即选取涨幅最高的n_top个股票', layout=self.label_layout)
        direction_top_label3 = widgets.Label(u'可设置为负：即选取跌幅最高的n_top个股票', layout=self.label_layout)
        self.direction_top = widgets.Dropdown(
            options={u'正(涨幅)': 1, u'负(跌幅)': -1},
            value=1,
            description=u'选取方向:',
        )
        self.direction_top_box = widgets.VBox([direction_top_label1, direction_top_label2, direction_top_label3,
                                               self.direction_top])

        self.widget = widgets.VBox([self.description, self.n_top_box, self.direction_top_box,
                                    self.xd_box, self.reversed_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def seed_choice_symbol_key(self):
        """返回更新股池中的种子symbol到具体策略中需要的策略关键字定义"""
        return 'symbol_pool'

    def make_pick_stock_unique(self):
        """对应按钮添加AbuPickStockNTop策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuPickStockNTop,
                       'n_top': self.n_top.value,
                       'direction_top': self.direction_top.value,
                       'reversed': self.reversed.value,
                       'xd': self.xd.value}

        factor_desc_key = u'涨跌幅选股n_top:{},方向:{},xd:{},反转:{}'.format(
            self.n_top.value, self.direction_top.value, self.xd.value, self.reversed.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuPickStockNTop"""
        return AbuPickStockNTop
